import torch
import os
from .train import traindata
from .test import test
from .models import *
from .prepare_data import prepare_data
from ray import tune
from ray.tune.schedulers import HyperBandScheduler
from ray.tune import CLIReporter
from functools import partial

import sys
sys.path.append("../")
from .utils import seed_everything, print_metrics

def hyperparameter_tuning(device, config):
    """
    Hyperparameter tuning for the deep learning model.
    :param device: The device to use for the model.
    :param config: The configuration for the model.
    """

    cpus_per_trial = config['fixed_vals']['cpus_per_trial']
    gpus_per_trial = config['fixed_vals']['gpus_per_trial']
    num_samples = config['fixed_vals']['num_samples']
    epochs = config['fixed_vals']['epochs']
    model_label = config['combination']['model_label']
    data_dir = config['combination']['data_dir']
    mode = config['combination']['mode']
    class_weights = config['combination']['class_weights']
    kmer_one_hot = config['fixed_vals']['kmer_one_hot']
    output_size = config['fixed_vals']['output_size']

    seed_everything()

    # ------------------------------------------------------------------------------------------

    scheduler = HyperBandScheduler(
        metric="loss",
        mode="min",
        max_t=epochs,
        reduction_factor=2
    )

    reporter = CLIReporter(
        metric_columns=["loss", "accuracy", "training_iteration", 'mcc']
    )

    result = tune.run(
        partial(
            prepare_and_train,
            device=device,
            config_from_json=config
        ),
        resources_per_trial={"cpu": cpus_per_trial, "gpu": gpus_per_trial},
        config=config['hyperparameter_search_space'],
        num_samples=num_samples,
        scheduler=scheduler,
        progress_reporter=reporter
    )

    best_trial = result.get_best_trial('mcc', 'max', 'last')
    print("Best trial config:", best_trial.config)
    print("Best trial final validation loss:", best_trial.last_result["loss"])
    print("Best trial final validation accuracy:", best_trial.last_result["accuracy"])
    print("Best trial final validation mcc:", best_trial.last_result["mcc"])

    _, testloader, _, input_size, sequence_length = prepare_data(
        data_dir=data_dir,
        mode=mode,
        batch_size=best_trial.config['batch_size'],
        k=kmer_one_hot,
    )
    
    if model_label == 'mlp':
        best_trained_model = MLP(input_size, best_trial.config['hidden_size'], output_size, best_trial.config['dropout'])
    elif model_label == 'cnn':
        best_trained_model = CNN(input_size, best_trial.config['hidden_size'], output_size, best_trial.config['dropout'], sequence_length)
    elif model_label == 'lstm':
        best_trained_model = LSTM(input_size, best_trial.config['hidden_size'], False, best_trial.config['num_layers'], output_size, sequence_length, best_trial.config['dropout'], device)
    elif model_label == 'bi_lstm':
        best_trained_model = LSTM(input_size, best_trial.config['hidden_size'], True, best_trial.config['num_layers'], output_size, sequence_length, best_trial.config['dropout'], device)
    elif model_label == 'gru':
        best_trained_model = GRU(input_size, best_trial.config['hidden_size'], False, best_trial.config['num_layers'], output_size, sequence_length, best_trial.config['dropout'], device)
    elif model_label == 'bi_gru':
        best_trained_model = GRU(input_size, best_trial.config['hidden_size'], True, best_trial.config['num_layers'], output_size, sequence_length, best_trial.config['dropout'], device)
    elif model_label == 'cnn_lstm':
        best_trained_model = CNN_LSTM(input_size, best_trial.config['hidden_size'], False, best_trial.config['num_layers'], sequence_length, output_size, best_trial.config['dropout'], device)
    elif model_label == 'cnn_bi_lstm':
        best_trained_model = CNN_LSTM(input_size, best_trial.config['hidden_size'], True, best_trial.config['num_layers'], sequence_length, output_size, best_trial.config['dropout'], device)
    elif model_label == 'cnn_gru':
        best_trained_model = CNN_GRU(input_size, best_trial.config['hidden_size'], False, best_trial.config['num_layers'], sequence_length, output_size, best_trial.config['dropout'], device)
    elif model_label == 'cnn_bi_gru':
        best_trained_model = CNN_GRU(input_size, best_trial.config['hidden_size'], True, best_trial.config['num_layers'], sequence_length, output_size, best_trial.config['dropout'], device)
    else:
        raise ValueError('Model label not implemented', model_label)
        

    best_trained_model.to(device)

    best_checkpoint_dir = best_trial.checkpoint.value
    model_state, optimizer_state = torch.load(os.path.join(
        best_checkpoint_dir, "checkpoint"))
    best_trained_model.load_state_dict(model_state)

    metrics = test(device, best_trained_model, testloader)
    print_metrics(model_label, mode, data_dir, kmer_one_hot, class_weights, metrics)

def prepare_and_train(config, device, config_from_json):
    
    data_dir = config_from_json['combination']['data_dir']
    mode = config_from_json['combination']['mode']
    kmer_one_hot = config_from_json['fixed_vals']['kmer_one_hot']
    batch_size = config['batch_size']
    
    trainloader, _, validloader, input_size, sequence_length = prepare_data(
        data_dir=data_dir,
        mode=mode,
        batch_size=batch_size,
        k=kmer_one_hot,
    )
    
    traindata(config, device, config_from_json, trainloader, validloader, input_size, sequence_length)


"""
 This module is used for obtaining the properties of amino acids or their pairs

from the aaindex database. You can freely use and distribute it. If you hava

any problem, you could contact with us timely!

Authors: Dongsheng Cao and Yizeng Liang.

Date: 2012.09.10

Email: oriental-cds@163.com
Altered and converted to python 3.6 by Ana Marta Sequeira 05/2019
"""
import sys, os

AALetter=["A","R","N","D","C","E","Q","G","H","I","L","K","M","F","P","S","T","W","Y","V"]
_aaindex = dict()

#####################################################################################################


class Record:
	"""
	Amino acid index (AAindex) Record
	"""
	aakeys = 'ARNDCQEGHILKMFPSTWYV'
	def __init__(self):
		self.key = None
		self.desc = ''
		self.ref = ''
		self.authors = ''
		self.title = ''
		self.journal = ''
		self.correlated = dict()
		self.index = dict()
		self.comment = ''

	def extend(self, row):
		i = len(self.index)
		for x in row:
			self.index[self.aakeys[i]] = x
			i += 1

	def get(self, aai, aaj=None, d=None):
		assert aaj is None
		return self.index.get(aai, d)

	def __getitem__(self, aai):
		return self.get(aai)

	def median(self):
		x = sorted([_f for _f in list(self.index.values()) if _f])
		if len(x) % 2 == 1:
			return x[len(x)//2]
		return (sum(x[len(x)//2-1:len(x)//2+1]))/2

	def __str__(self):
		desc = self.desc.replace('\n', ' ').strip()
		return '%s(%s: %s)' % (self.__class__.__name__, self.key, desc)
		
#####################################################################################################


class MatrixRecord(Record):
	"""
	Matrix record for mutation matrices or pair-wise contact potentials
	"""
	def __init__(self):
		Record.__init__(self)
		self.index = []
		self.rows = dict()
		self.cols = dict()

	def extend(self, row):
		self.index.append(row)

	def _get(self, aai, aaj):
		i = self.rows[aai]
		j = self.cols[aaj]
		return self.index[i][j]

	def get(self, aai, aaj, d=None):
		try:
			return self._get(aai, aaj)
		except:
			pass
		try:
			return self._get(aaj, aai)
		except:
			return d

	def __getitem__(self, aaij):
		return self.get(aaij[0], aaij[1])

	def median(self):
		x = []
		for y in self.index:
			x.extend([_f for _f in y if _f])
		x.sort()
		if len(x) % 2 == 1:
			return x[len(x)//2]
		return sum(x[len(x)/2-1:len(x)/2+1])/2.0
#####################################################################################################


def search(pattern, searchtitle=True, casesensitive=False):
	"""
	Search for pattern in description and title (optional) of all records and
	return matched records as list. By default search case insensitive.

	:param pattern:
	:param searchtitle:
	:param casesensitive:
	:return:
	"""
	whatcase = lambda i: i
	if not casesensitive:
		pattern = pattern.lower()
		whatcase = lambda i: i.lower()
	matches = []
	for record in _aaindex.values():
		if pattern in whatcase(record.desc) or searchtitle and pattern in whatcase(record.title):
			matches.append(record)
	return matches

#####################################################################################################


def grep(pattern):
	"""
	Search for pattern in title and description of all records (case insensitive) and print results on standard output.

	:param pattern:
	:return:
	"""
	for record in search(pattern):
		print(record)
#####################################################################################################


def get(key):
	"""
	Get record for key

	:param key:
	:return:
	"""
	if len(_aaindex) == 0:
		init()
	return _aaindex[key]

#####################################################################################################


def _float_or_none(x):
	if x == 'NA' or x == '-':
		return None
	return float(x)

#####################################################################################################


def init(path=None, index='123'):
	"""
	Read in the aaindex files. You need to run this (once) before you can
	access any records. If the files are not within the current directory,
	you need to specify the correct directory path. By default all three
	aaindex files are read in.

	:param path:
	:param index:
	:return:
	"""
	index = str(index)
	if path is None:
		for path in [os.path.split(__file__)[0], '.']:
			if os.path.exists(os.path.join(path, 'aaindex' + index[0])):
				break
		print('path =', path, file=sys.stderr)
	if '1' in index:
		_parse(path + '/aaindex1', Record)
	if '2' in index:
		_parse(path + '/aaindex2', MatrixRecord)
	if '3' in index:
		_parse(path + '/aaindex3', MatrixRecord)
#####################################################################################################


def init_from_file(filename, type=Record):
	_parse(filename, type)
#####################################################################################################


def _parse(filename, rec, quiet=True):
	"""
	Parse aaindex input file. `rec` must be `Record` for aaindex1 and `MarixRecord` for aaindex2 and aaindex3.

	:param filename:
	:param rec:
	:param quiet:
	:return:
	"""
	if not os.path.exists(filename):
		import urllib.request, urllib.parse, urllib.error
		url = 'ftp://ftp.genome.jp/pub/db/community/aaindex/' + os.path.split(filename)[1]
#		print 'Downloading "%s"' % (url)
		filename = urllib.request.urlretrieve(url, filename)[0]
#		print 'Saved to "%s"' % (filename)
	f = open(filename)

	current = rec()
	lastkey = None
	for line in f:
		key = line[0:2]
		if key[0] == ' ':
			key = lastkey
		if key == '//':
			_aaindex[current.key] = current
			current = rec()
		elif key == 'H ':
			current.key = line[2:].strip()
		elif key == 'R ':
			current.ref += line[2:]
		elif key == 'D ':
			current.desc += line[2:]
		elif key == 'A ':
			current.authors += line[2:]
		elif key == 'T ':
			current.title += line[2:]
		elif key == 'J ':
			current.journal += line[2:]
		elif key == '* ':
			current.comment += line[2:]
		elif key == 'C ':
			a = line[2:].split()
			for i in range(0, len(a), 2):
				current.correlated[a[i]] = float(a[i+1])
		elif key == 'I ':
			a = line[1:].split()
			if a[0] != 'A/L':
				current.extend(list(map(_float_or_none, a)))
			elif list(Record.aakeys) != [i[0] for i in a] + [i[-1] for i in a]:
				print('Warning: wrong amino acid sequence for', current.key)
			else:
				try:
					assert list(Record.aakeys[:10]) == [i[0] for i in a]
					assert list(Record.aakeys[10:]) == [i[2] for i in a]
				except:
					print('Warning: wrong amino acid sequence for', current.key)
		elif key =='M ':
			a = line[2:].split()
			if a[0] == 'rows':
				if a[4] == 'rows':
					a.pop(4)
				assert a[3] == 'cols' and len(a) == 6
				i = 0
				for aa in a[2]:
					current.rows[aa] = i
					i += 1
				i = 0
				for aa in a[5]:
					current.cols[aa] = i
					i += 1
			else:
				current.extend(list(map(_float_or_none, a)))
		elif not quiet:
			print('Warning: line starts with "%s"' % (key))
		lastkey = key
	f.close()
#####################################################################################################


def get_aa_index1(name, path='.'):
	"""
	Get the amino acid property values from aaindex1

	:param name: name is the name of amino acid property (e.g., KRIW790103)
	:param path:
	:return: get_aa_index1(name). result is a dict form containing the properties of 20 amino acids
	"""
	init(path=path)
	name=str(name)
	temp=get(str.strip(name))
	res={}
	for i in AALetter:
		res[i]=temp.get(i)
	return res

#####################################################################################################


def get_aa_index23(name, path='.'):
	"""
	Get the amino acid property values from aaindex2 and aaindex3

	:param name: name is the name of amino acid property (e.g.,TANS760101,GRAR740104)
	:param path:
	:return: get_aa_index23(name). result is a dict form containing the properties of 400 amino acid pairs
	"""
	init(path=path)
	name=str(name)
	temp=get(str.strip(name))
	res={}
	for i in AALetter:
		for j in AALetter:
			res[i+j]=temp.get(i,j)
	return res
#####################################################################################################


if __name__ == "__main__":

	#	init(path='.')
	#	grep('volume')
	#	x = get('KRIW790103')
	#	print x
	#	print x.get('W')
	temp1=get_aa_index1('KRIW790103')
	print(len(temp1))
	
	temp2=get_aa_index23('TANS760101')
	print(len(temp2))
	temp2=get_aa_index23('GRAR740104')
	print(len(temp2))




# -*- coding: utf-8 -*-
"""
This module is used for computing the Autocorrelation based different

 properties of AADs.You can also input your properties of AADs, then it can help you

to compute Autocorrelation based on the property of AADs. Currently, You

can get 720  for a given protein sequence based on our provided physicochemical

properties of AADs. You can freely use and distribute it. If you hava  any problem,

you could contact with us timely!

References:

[1]: http://www.genome.ad.jp/dbget/aaindex.html

[2]:Feng, Z.P. and Zhang, C.T. (2000) Prediction of membrane protein types based on

the hydrophobic index of amino acids. J Protein Chem, 19, 269-275.

[3]:Horne, D.S. (1988) Prediction of protein helix content from an autocorrelation

analysis of sequence hydrophobicities. Biopolymers, 27, 451-477.

[4]:Sokal, R.R. and Thomson, B.A. (2006) Population structure inferred by local

spatial autocorrelation: an Usage from an Amerindian tribal population. Am J

Phys Anthropol, 129, 121-131.

Authors: Dongsheng Cao and Yizeng Liang.

Date: 2010.11.22

Email: oriental-cds@163.com

Altered to python 3.6 by Ana Marta Sequeira 05/2019
"""

import math


AALetter=["A","R","N","D","C","Q","E","G","H","I","L","K","M","F","P","S","T","W","Y","V"]


_hydrophobicity = {"A":0.02, "R":-0.42, "N":-0.77, "D":-1.04, "C":0.77, "Q":-1.10, "E":-1.14, "G":-0.80, "H":0.26, "I":1.81, "L":1.14, "K":-0.41, "M":1.00, "F":1.35, "P":-0.09, "S":-0.97, "T":-0.77, "W":1.71, "Y":1.11, "V":1.13}

_av_flexibility = {"A":0.357, "R":0.529, "N":0.463, "D":0.511, "C":0.346, "Q":0.493, "E":0.497, "G":0.544, "H":0.323, "I":0.462, "L":0.365, "K":0.466, "M":0.295, "F":0.314, "P":0.509, "S":0.507, "T":0.444, "W":0.305, "Y":0.420, "V":0.386}

_polarizability = {"A":0.046, "R":0.291, "N":0.134, "D":0.105, "C":0.128, "Q":0.180, "E":0.151, "G":0.000, "H":0.230, "I":0.186, "L":0.186, "K":0.219, "M":0.221, "F":0.290, "P":0.131, "S":0.062, "T":0.108, "W":0.409, "Y":0.298, "V":0.140}

_free_energy = {"A":-0.368, "R":-1.03, "N":0.0, "D":2.06, "C":4.53, "Q":0.731, "E":1.77, "G":-0.525, "H":0.0, "I":0.791, "L":1.07, "K":0.0, "M":0.656, "F":1.06, "P":-2.24, "S":-0.524, "T":0.0, "W":1.60, "Y":4.91, "V":0.401}

_residue_asa = {"A":115.0, "R":225.0, "N":160.0, "D":150.0, "C":135.0, "Q":180.0, "E":190.0, "G":75.0, "H":195.0, "I":175.0, "L":170.0, "K":200.0, "M":185.0, "F":210.0, "P":145.0, "S":115.0, "T":140.0, "W":255.0, "Y":230.0, "V":155.0}

_residue_vol = {"A":52.6, "R":109.1, "N":75.7, "D":68.4, "C":68.3, "Q":89.7, "E":84.7, "G":36.3, "H":91.9, "I":102.0, "L":102.0, "K":105.1, "M":97.7, "F":113.9, "P":73.6, "S":54.9, "T":71.2, "W":135.4, "Y":116.2, "V":85.1}

_steric = {"A":0.52, "R":0.68, "N":0.76, "D":0.76, "C":0.62, "Q":0.68, "E":0.68, "G":0.00, "H":0.70, "I":1.02, "L":0.98, "K":0.68, "M":0.78, "F":0.70, "P":0.36, "S":0.53, "T":0.50, "W":0.70, "Y":0.70, "V":0.76}

_mutability = {"A":100.0, "R":65.0, "N":134.0, "D":106.0, "C":20.0, "Q":93.0, "E":102.0, "G":49.0, "H":66.0, "I":96.0, "L":40.0, "K":-56.0, "M":94.0, "F":41.0, "P":56.0, "S":120.0, "T":97.0, "W":18.0, "Y":41.0, "V":74.0}


###You can continuely add other properties of AADs to compute the pydpi_py3 of protein sequence.


_aa_property=(_hydrophobicity, _av_flexibility, _polarizability, _free_energy, _residue_asa, _residue_vol, _steric, _mutability)

_aa_property_name=('_Hydrophobicity', '_AvFlexibility', '_Polarizability', '_FreeEnergy', '_ResidueASA', '_ResidueVol', '_Steric', '_Mutability')

##################################################################################################

def _mean(listvalue):
	"""
	The mean value of the list data.
	:param listvalue:
	:return:
	"""

	return sum(listvalue)/len(listvalue)
##################################################################################################


def _std(listvalue, ddof=1):
	"""
	The standard deviation of the list data.
	"""
	mean=_mean(listvalue)
	temp=[math.pow(i-mean,2) for i in listvalue]
	res=math.sqrt(sum(temp)/(len(listvalue)-ddof))
	return res
##################################################################################################


def normalize_each_aap(aap):
	"""
	All of the amino acid indices are centralized and standardized before the calculation.

	:param aap: AAP is a dict form containing the properties of 20 amino acids
	:return: result is the a dict form containing the normalized properties of 20 amino acids.
	"""

	result={}
	if len(list(aap.values()))!=20:
		print('You can not input the correct number of properities of Amino acids!')
	else:
		result={}
		for i,j in list(aap.items()):
			result[i]= (j - _mean(list(aap.values()))) / _std(list(aap.values()), ddof=0)

	return result


def calculate_each_normalized_moreau_broto_auto(protein_sequence, aap, aap_name):
	"""
	you can use the function to compute MoreauBrotoAuto for different properties based on AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:param aap: AAP is a dict form containing the properties of 20 amino acids (e.g., _AvFlexibility).
	:param aap_name: AAPName is a string used for indicating the property (e.g., '_AvFlexibility')
	:return: result is a dict form containing 30 Normalized Moreau-Broto autocorrelation based on the given property.
	"""
	aa_pdic = normalize_each_aap(aap)

	result={}
	for i in range(1,31):
		temp=0
		for j in range(len(protein_sequence) - i):
			temp= temp + aa_pdic[protein_sequence[j]] * aa_pdic[protein_sequence[j + 1]]
		if len(protein_sequence)-i==0:
			result['MoreauBrotoAuto' + aap_name + str(i)]=round(temp / (len(protein_sequence)), 3)
		else:
			result['MoreauBrotoAuto' + aap_name + str(i)]=round(temp / (len(protein_sequence) - i), 3)

	return result


def calculate_each_moran_auto(protein_sequence, aap, aap_name):
	"""
	you can use the function to compute MoranAuto for different properties based on AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:param aap: AAP is a dict form containing the properties of 20 amino acids (e.g., _AvFlexibility).
	:param aap_name: AAPName is a string used for indicating the property (e.g., '_AvFlexibility').
	:return: result is a dict form containing 30 Moran autocorrelation based on the given property.
	"""

	aa_pdic=normalize_each_aap(aap)

	cds=0
	for i in AALetter:
		cds= cds + (protein_sequence.count(i)) * (aa_pdic[i])
	Pmean=cds/len(protein_sequence)

	cc=[]
	for i in protein_sequence:
		cc.append(aa_pdic[i])

	k=(_std(cc,ddof=0))**2

	result={}
	for i in range(1,31):
		temp=0
		for j in range(len(protein_sequence) - i):
				
			temp= temp + (aa_pdic[protein_sequence[j]] - Pmean) * (aa_pdic[protein_sequence[j + i]] - Pmean)
		if len(protein_sequence)-i == 0:
			result['MoranAuto' + aap_name + str(i)]=round(temp / (len(protein_sequence)) / k, 3)
		else:
			result['MoranAuto' + aap_name + str(i)]=round(temp / (len(protein_sequence) - i) / k, 3)

	return result


def calculate_each_geary_auto(protein_sequence, aap, aap_name):
	"""
	you can use the function to compute GearyAuto for different properties based on AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:param aap: AAP is a dict form containing the properties of 20 amino acids (e.g., _AvFlexibility).
	:param aap_name: AAPName is a string used for indicating the property (e.g., '_AvFlexibility').
	:return: result is a dict form containing 30 Geary autocorrelation based on the given property.
	"""

	aa_pdic=normalize_each_aap(aap)

	cc=[]
	for i in protein_sequence:
		cc.append(aa_pdic[i])

	K= ((_std(cc))**2) * len(protein_sequence) / (len(protein_sequence) - 1)
	Result={}
	for i in range(1,31):
		temp=0
		for j in range(len(protein_sequence) - i):
				
			temp= temp + (aa_pdic[protein_sequence[j]] - aa_pdic[protein_sequence[j + i]]) ** 2
		if len(protein_sequence)-i==0:
			Result['GearyAuto' + aap_name + str(i)]=round(temp / (2 * (len(protein_sequence))) / K, 3)
		else:
			Result['GearyAuto' + aap_name + str(i)]=round(temp / (2 * (len(protein_sequence) - i)) / K, 3)
	return Result
##################################################################################################


def calculate_normalized_moreau_broto_auto(protein_sequence, aa_property, aa_property_name):
	"""
	A method used for computing MoreauBrotoAuto for all properties.

	:param protein_sequence: protein is a pure protein sequence.
	:param aa_property: AAProperty is a list or tuple form containing the properties of 20 amino acids (e.g., _AAProperty).
	:param aa_property_name: AAPName is a list or tuple form used for indicating the property (e.g., '_AAPropertyName').
	:return: result is a dict form containing 30*p Normalized Moreau-Broto autocorrelation based on the given properties.
	"""
	result = {}
	for i in range(len(aa_property)):
		result[aa_property_name[i]]=calculate_each_normalized_moreau_broto_auto(protein_sequence, aa_property[i], aa_property_name[i])
	return result


def calculate_moran_auto(protein_sequence, aa_property, aa_property_name):
	"""
	A method used for computing MoranAuto for all properties

	:param protein_sequence: protein is a pure protein sequence.
	:param aa_property: AAProperty is a list or tuple form containing the properties of 20 amino acids (e.g., _AAProperty).
	:param aa_property_name: AAPName is a list or tuple form used for indicating the property (e.g., '_AAPropertyName').
	:return: result is a dict form containing 30*p Moran autocorrelation based on the given properties.
	"""

	result={}
	for i in range(len(aa_property)):
		result[aa_property_name[i]]=calculate_each_moran_auto(protein_sequence, aa_property[i], aa_property_name[i])

	return result


def calculate_geary_auto(protein_sequence, aa_property, aa_property_name):
	"""
	A method used for computing GearyAuto for all properties

	:param protein_sequence: protein is a pure protein sequence
	:param aa_property: AAProperty is a list or tuple form containing the properties of 20 amino acids (e.g., _AAProperty).
	:param aa_property_name: AAPName is a list or tuple form used for indicating the property (e.g., '_AAPropertyName').
	:return: result is a dict form containing 30*p Geary autocorrelation based on the given properties.
	"""

	result={}
	for i in range(len(aa_property)):
		result[aa_property_name[i]]=calculate_each_geary_auto(protein_sequence, aa_property[i], aa_property_name[i])

	return result

########################NormalizedMoreauBorto##################################


def calculate_normalized_moreau_broto_auto_hydrophobicity(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation based on hydrophobicity.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on Hydrophobicity.
	"""

	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _hydrophobicity, '_Hydrophobicity')
	return result


def calculate_normalized_moreau_broto_auto_av_flexibility(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation based on AvFlexibility.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on AvFlexibility.
	"""
	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _av_flexibility, '_AvFlexibility')
	return result


def calculate_normalized_moreau_broto_auto_polarizability(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation based on Polarizability.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on Polarizability.
	"""

	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _polarizability, '_Polarizability')
	return result


def calculate_normalized_moreau_broto_auto_free_energy(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation  based on FreeEnergy.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on FreeEnergy.
	"""
	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _free_energy, '_FreeEnergy')
	return result



def calculate_normalized_moreau_broto_auto_residue_asa(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation based on ResidueASA.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on ResidueASA.
	"""
	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _residue_asa, '_ResidueASA')
	return result


def calculate_normalized_moreau_broto_auto_residue_vol(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation based on ResidueVol.

	:param protein_sequence: protein is a pure protein sequence.
	:return:  result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on ResidueVol.
	"""
	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _residue_vol, '_ResidueVol')
	return result


def calculate_normalized_moreau_broto_auto_steric(protein_sequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation  based on Steric.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on Steric.
	"""
	result=calculate_each_normalized_moreau_broto_auto(protein_sequence, _steric, '_Steric')
	return result


def CalculateNormalizedMoreauBrotoAutoMutability(ProteinSequence):
	"""
	Calculate the NormalizedMoreauBorto Autocorrelation based on Mutability.

	:param ProteinSequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Normalized Moreau-Broto Autocorrelation based on Mutability.
	"""
	result=calculate_each_normalized_moreau_broto_auto(ProteinSequence, _mutability, '_Mutability')
	return result


##############################MoranAuto######################################

def calculate_moran_auto_hydrophobicity(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation  based on hydrophobicity.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on hydrophobicity.
	"""
	result=calculate_each_moran_auto(protein_sequence, _hydrophobicity, '_Hydrophobicity')
	return result
	

def calculate_moran_auto_av_flexibility(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on	AvFlexibility.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on AvFlexibility.
	"""
	result=calculate_each_moran_auto(protein_sequence, _av_flexibility, '_AvFlexibility')
	return result


def calculate_moran_auto_polarizability(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on Polarizability.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on Polarizability.
	"""
	result=calculate_each_moran_auto(protein_sequence, _polarizability, '_Polarizability')
	return result


def calculate_moran_auto_free_energy(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on FreeEnergy.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on FreeEnergy.
	"""
	result=calculate_each_moran_auto(protein_sequence, _free_energy, '_FreeEnergy')
	return result


def calculate_moran_auto_residue_asa(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on ResidueASA.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on ResidueASA.
	"""
	result=calculate_each_moran_auto(protein_sequence, _residue_asa, '_ResidueASA')
	return result


def calculate_moran_auto_residue_vol(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on ResidueVol.

	:param protein_sequence:  protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on ResidueVol.
	"""
	result=calculate_each_moran_auto(protein_sequence, _residue_vol, '_ResidueVol')
	return result


def calculate_moran_auto_steric(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on AutoSteric.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on AutoSteric.
	"""
	result=calculate_each_moran_auto(protein_sequence, _steric, '_Steric')
	return result


def calculate_moran_auto_mutability(protein_sequence):
	"""
	Calculate the MoranAuto Autocorrelation based on Mutability.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Moran Autocorrelation based on Mutability.
	"""
	result=calculate_each_moran_auto(protein_sequence, _mutability, '_Mutability')
	return result

################################GearyAuto#####################################


def calculate_geary_auto_hydrophobicity(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on hydrophobicity.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on hydrophobicity.
	"""
	result=calculate_each_geary_auto(protein_sequence, _hydrophobicity, '_Hydrophobicity')
	return result
	

def calculate_geary_auto_av_flexibility(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on AvFlexibility.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on AvFlexibility.
	"""
	result=calculate_each_geary_auto(protein_sequence, _av_flexibility, '_AvFlexibility')
	return result


def calculate_geary_auto_polarizability(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on Polarizability.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on Polarizability.
	"""
	result=calculate_each_geary_auto(protein_sequence, _polarizability, '_Polarizability')
	return result


def calculate_geary_auto_free_energy(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation based on FreeEnergy.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on FreeEnergy.
	"""
	result=calculate_each_geary_auto(protein_sequence, _free_energy, '_FreeEnergy')
	return result


def calculate_geary_auto_residue_asa(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on ResidueASA.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on ResidueASA.
	"""
	result = calculate_each_geary_auto(protein_sequence, _residue_asa, '_ResidueASA')
	return result


def calculate_geary_auto_residue_vol(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on ResidueVol.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on ResidueVol.
	"""
	result = calculate_each_geary_auto(protein_sequence, _residue_vol, '_ResidueVol')
	return result


def calculate_geary_auto_steric(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on Steric.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on Steric.
	"""
	result=calculate_each_geary_auto(protein_sequence, _steric, '_Steric')
	return result


def calculate_geary_auto_mutability(protein_sequence):
	"""
	Calculate the GearyAuto Autocorrelation  based on Mutability.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30 Geary Autocorrelation based on Mutability.
	"""
	result=calculate_each_geary_auto(protein_sequence, _mutability, '_Mutability')
	return result
##################################################################################################


def calculate_normalized_moreau_broto_auto_total(protein_sequence):
	"""
	A method used for computing normalized Moreau Broto autocorrelation  based on 8 proterties of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return:
		result is a dict form containing 30*8=240 normalized Moreau Broto autocorrelation  based on the given
		properties(i.e., _AAPropert).
	"""
	result={}
	result.update(calculate_normalized_moreau_broto_auto_hydrophobicity(protein_sequence))
	result.update(calculate_normalized_moreau_broto_auto_av_flexibility(protein_sequence))
	result.update(calculate_normalized_moreau_broto_auto_polarizability(protein_sequence))
	result.update(calculate_normalized_moreau_broto_auto_free_energy(protein_sequence))
	result.update(calculate_normalized_moreau_broto_auto_residue_asa(protein_sequence))
	result.update(calculate_normalized_moreau_broto_auto_residue_vol(protein_sequence))
	result.update(calculate_normalized_moreau_broto_auto_steric(protein_sequence))
	result.update(CalculateNormalizedMoreauBrotoAutoMutability(protein_sequence))
	return result

def calculate_moran_auto_total(protein_sequence):
	"""
	A method used for computing Moran autocorrelation based on 8 properties of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return: result is a dict form containing 30*8=240 Moran autocorrelation  based on the given properties(i.e., _AAPropert).
	"""
	result={}
	result.update(calculate_moran_auto_hydrophobicity(protein_sequence))
	result.update(calculate_moran_auto_av_flexibility(protein_sequence))
	result.update(calculate_moran_auto_polarizability(protein_sequence))
	result.update(calculate_moran_auto_free_energy(protein_sequence))
	result.update(calculate_moran_auto_residue_asa(protein_sequence))
	result.update(calculate_moran_auto_residue_vol(protein_sequence))
	result.update(calculate_moran_auto_steric(protein_sequence))
	result.update(calculate_moran_auto_mutability(protein_sequence))
	return result

def calculate_geary_auto_total(protein_sequence):
	"""
	A method used for computing Geary autocorrelation based on 8 properties of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return:
		result is a dict form containing 30*8=240 Geary autocorrelation based on the given
		properties(i.e., _AAPropert).
	"""
	result={}
	result.update(calculate_geary_auto_hydrophobicity(protein_sequence))
	result.update(calculate_geary_auto_av_flexibility(protein_sequence))
	result.update(calculate_geary_auto_polarizability(protein_sequence))
	result.update(calculate_geary_auto_free_energy(protein_sequence))
	result.update(calculate_geary_auto_residue_asa(protein_sequence))
	result.update(calculate_geary_auto_residue_vol(protein_sequence))
	result.update(calculate_geary_auto_steric(protein_sequence))
	result.update(calculate_geary_auto_mutability(protein_sequence))
	return result

##################################################################################################


def calculate_auto_total(protein_sequence):
	"""
	A method used for computing all autocorrelation  based on 8 properties of AADs.

	:param protein_sequence: protein is a pure protein sequence.
	:return:
		result is a dict form containing 30*8*3=720 normalized Moreau Broto, Moran, and Geary autocorrelation
		based on the given properties(i.e., _AAPropert).
	"""
	result={}
	result.update(calculate_normalized_moreau_broto_auto_total(protein_sequence))
	result.update(calculate_moran_auto_total(protein_sequence))
	result.update(calculate_geary_auto_total(protein_sequence))
	return result


if __name__ == "__main__":
	protein="ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
	temp1=calculate_normalized_moreau_broto_auto(protein, aa_property=_aa_property, aa_property_name=_aa_property_name)
	#print temp1
	temp2=calculate_moran_auto_mutability(protein)
	print(temp2)
	print(len(calculate_auto_total(protein)))





import numpy as np
import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim import Adam, SGD
from .test import test
from .models import *
from .prepare_data import prepare_data
from ray import tune
import os

import sys
sys.path.append("../")
from .utils import seed_everything


def traindata(config, device, config_from_json, trainloader, validloader, input_size, sequence_length, checkpoint_dir=None):
    """
    Train the model for a number of epochs.
    :param config: Dictionary of hyperparameters to be tuned.
    :param device: Device to be used for training.
    :param fixed_vals: Dictionary of fixed parameters.
    :param checkpoint_dir: Directory to save the model.
    """
    
    seed_everything()
    
    # Fixed values
    do_tuning = config_from_json['do_tuning']
    model_label = config_from_json['combination']['model_label']
    output_size = config_from_json['fixed_vals']['output_size']
    optimizer_label = config_from_json['fixed_vals']['optimizer_label']
    epochs = config_from_json['fixed_vals']['epochs']
    patience = config_from_json['fixed_vals']['patience']
    loss_function = config_from_json['fixed_vals']['loss_function']
    last_loss = 100

    # Hyperparameters to tune
    hidden_size = config['hidden_size']
    dropout = config['dropout']
    lr = config['lr']
    num_layers = config['num_layers']

    if model_label == 'mlp':
        model = MLP(input_size, hidden_size, output_size, dropout).to(device)
    elif model_label == 'cnn':
        model = CNN(input_size, hidden_size, output_size, dropout, sequence_length).to(device)
    elif model_label == 'lstm':
        model = LSTM(input_size, hidden_size, False, num_layers, output_size, sequence_length, dropout, device).to(device)
    elif model_label == 'bi_lstm':
        model = LSTM(input_size, hidden_size, True, num_layers, output_size, sequence_length, dropout, device).to(device)
    elif model_label == 'gru':
        model = GRU(input_size, hidden_size, False, num_layers, output_size, sequence_length, dropout, device).to(device)
    elif model_label == 'bi_gru':
        model = GRU(input_size, hidden_size, True, num_layers, output_size, sequence_length, dropout, device).to(device)
    elif model_label == 'cnn_lstm':
        model = CNN_LSTM(input_size, hidden_size, False, num_layers, sequence_length, output_size, dropout, device).to(device)
    elif model_label == 'cnn_bi_lstm':
        model = CNN_LSTM(input_size, hidden_size, True, num_layers, sequence_length, output_size, dropout, device).to(device)
    elif model_label == 'cnn_gru':
        model = CNN_GRU(input_size, hidden_size, False, num_layers, sequence_length, output_size, dropout, device).to(device)
    elif model_label == 'cnn_bi_gru':
        model = CNN_GRU(input_size, hidden_size, True, num_layers, sequence_length, output_size, dropout, device).to(device)
    else:
        raise ValueError('Model label not implemented', model_label)

    if(optimizer_label == 'adam'):
        optimizer = Adam(model.parameters(), lr=lr)
    elif(optimizer_label == 'sgd'):
        optimizer = SGD(model.parameters(), lr=lr)
    else:
        raise ValueError("optimizer_label must be either 'adam' or 'sgd'")

    scheduler = ReduceLROnPlateau(optimizer, 'min')

    # ------------------------------------------------------------------------------------------------

    if do_tuning and checkpoint_dir:
        model_state, optimizer_state = torch.load(
            os.path.join(checkpoint_dir, "checkpoint"))
        model.load_state_dict(model_state)
        optimizer.load_state_dict(optimizer_state)

    # ------------------------------------------------------------------------------------------------

    trigger_times = 0
    for epoch in range(1, epochs+1):
        model.train()

        for i, (inputs, targets) in enumerate(trainloader):
            inputs, targets = inputs.to(device), targets.to(device)

            # Zero the gradients
            optimizer.zero_grad()

            # Forward and backward propagation
            output = model(inputs)
            loss = loss_function(output, targets)
            loss.backward()
            optimizer.step()

            # Show progress
            if i % 100 == 0 or i == len(trainloader):
                print(f'[{epoch}/{epochs}, {i}/{len(trainloader)}] loss: {loss.item():.8}')

        # Early stopping
        current_loss, val_acc, val_mcc = validation(model, device, validloader, loss_function)
        print('The Current Loss:', current_loss)

        if current_loss >= last_loss or torch.isnan(loss):
            trigger_times += 1
            if torch.isnan(loss):
                print('NAN loss! trigger_times:', trigger_times)
            else:
                print('trigger Times:', trigger_times)

            if trigger_times >= patience:
                print('Early stopping!\nStart to test process.')
                if do_tuning:
                    with tune.checkpoint_dir(epoch) as checkpoint_dir:
                        path = os.path.join(checkpoint_dir, "checkpoint")
                        torch.save((model.state_dict(), optimizer.state_dict()), path)
                    tune.report(loss=current_loss, accuracy=val_acc, mcc=val_mcc)
                    return
                else:
                    return model

        else:
            print('trigger times: 0')
            trigger_times = 0

        last_loss = current_loss
        scheduler.step(current_loss)
        if do_tuning:
            tune.report(loss=current_loss, accuracy=val_acc, mcc=val_mcc)
        
    # if do_tuning == False:
    return model
        

def validation(model, device, validloader, loss_function):
    """
    Validate the model.
    :param model: Model to be validated.
    :param device: Device to be used for validation.
    :param validloader: Data loader for validation.
    :param loss_function: Loss function to be used.
    :return: The loss, accuracy and mcc of the model.
    """
    model.eval()
    loss_total = 0

    with torch.no_grad():
        for (inputs, targets) in validloader:
            inputs, targets = inputs.to(device), targets.to(device)

            output = model(inputs)
            loss = loss_function(output, targets)
            loss_total += loss.item()

    metrics = test(device, model, validloader)

    return loss_total / len(validloader), metrics['accuracy'], metrics['mcc']



"""
##############################################################################
A class  used for computing different types of DNA encodings.
It contains encodings such one-hot-encoding. 
Authors: João Abreu
Date: 03/2022
Email:
##############################################################################
"""

import sys
from typing import Union
import numpy as np
sys.path.append('../')
from .utils import calculate_kmer_onehot, calculate_kmer_list

class DNAEncoder:
    def __init__(self, data: Union[str, np.ndarray]):
        
        if(isinstance(data, str)):
            self.dna_sequence = data.strip().upper()
            self.sequences = None
        else:
            self.sequences = data
            self.dna_sequence = None
              
    def one_hot_encode(self, dimension = 3):
        """
        From: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8138820/
        Calculates binary encoding. Each nucleotide is encoded by a four digit binary vector.
        :return: list with values of binary encoding
        """
        binary = {
            'A': [1, 0, 0, 0],
            'C': [0, 1, 0, 0],
            'G': [0, 0, 1, 0],
            'T': [0, 0, 0, 1],
            'N': [0, 0, 0, 0],
            1: [1, 0],
            0: [0, 1]            
        }
        
        binary2 = {
            'A': 1,
            'C': 2,
            'G': 3,
            'T': 4,
            'N': 0,
        }
        
        values = binary if dimension == 3 else binary2
        
        if(self.sequences is not None):
            return np.array([[values[i] for i in x] for x in self.sequences])
        elif(self.dna_sequence is not None):
            return np.array([values[i] for i in self.dna_sequence])
        else:
            print("Unexpected error: self.sequences and self.dna_sequence are None.")
            sys.exit(1)
            
    def chemical_encode(self):
        """
        From: https://academic.oup.com/bioinformatics/article/33/22/3518/4036387

        Calculates nucleotide chemical property

        Chemical property | Class	   | Nucleotides
        -------------------------------------------
        Ring structure 	  | Purine 	   | A, G
                          | Pyrimidine | C, T
        -------------------------------------------
        Hydrogen bond     | Weak 	   | A, T
         	              | Strong 	   | C, G
        -------------------------------------------
        Functional group  | Amino 	   | A, C
                          | Keto 	   | G, T

        :return: list with values of nucleotide chemical property
        """
        chemical_property = {
            'A': [1, 1, 1],
            'C': [0, 0, 1],
            'G': [1, 0, 0],
            'T': [0, 1, 0],
            'N': [0, 0, 0],
            1: [1, 0],
            0: [0, 1]
        }
        if(self.sequences is not None):
            return np.array([[chemical_property[i] for i in x] for x in self.sequences])
        elif(self.dna_sequence is not None):
            return np.array([chemical_property[i] for i in self.dna_sequence])
        else:
            print("Unexpected error: self.sequences and self.dna_sequence are None.")
            sys.exit(1)

    def kmer_one_hot_encode(self, k):
        """
        From: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8138820/
        Calculates binary encoding. Each nucleotide is encoded by a four digit binary vector.
        :return: list with values of binary encoding
        """
        if(self.sequences is not None):
            res = []
            d = calculate_kmer_onehot(k)
            for sequence in self.sequences:
                l = calculate_kmer_list(sequence, k)
                res.append(np.array([d[i] for i in l]))
            return np.array(res)
        elif(self.dna_sequence is not None):
            d = calculate_kmer_onehot(k)
            l = calculate_kmer_list(self.dna_sequence, k)
            return np.array([d[i] for i in l])
        else:
            print("Unexpected error: self.sequences and self.dna_sequence are None.")
            sys.exit(1)

if __name__ == "__main__":
    encoder = DNAEncoder("ACGT")
    print(encoder.one_hot_encode())
    print(encoder.chemical_encode())
    print(encoder.kmer_one_hot_encode(2))


# -*- coding: utf-8 -*-
"""
#########################################################################################

Instead of using the conventional 20-D amino acid composition to represent the sample

of a protein, Prof. Kuo-Chen Chou proposed the pseudo amino acid (PseAA) composition 

in order for inluding the sequence-order information. Based on the concept of Chou's 
 
pseudo amino acid composition, the server PseAA was designed in a flexible way, allowing 
 
users to generate various kinds of pseudo amino acid composition for a given protein
 
sequence by selecting different parameters and their combinations. This module aims at 
 
computing two types of PseAA pydpi_py3: Type I and Type II.
 
You can freely use and distribute it. If you have any problem, you could contact 
 
with us timely.

References:

[1]: Kuo-Chen Chou. Prediction of Protein Cellular Attributes Using Pseudo-Amino Acid 

Composition. PROTEINS: Structure, Function, and Genetics, 2001, 43: 246-255.

[2]: http://www.csbio.sjtu.edu.cn/bioinf/PseAAC/

[3]: http://www.csbio.sjtu.edu.cn/bioinf/PseAAC/type2.htm

[4]: Kuo-Chen Chou. Using amphiphilic pseudo amino acid composition to predict enzyme 

subfamily classes. Bioinformatics, 2005,21,10-19.

Authors: Dongsheng Cao and Yizeng Liang.

Date: 2012.9.2

Email: oriental-cds@163.com


The hydrophobicity values are from JACS, 1962, 84: 4240-4246. (C. Tanford).

The hydrophilicity values are from PNAS, 1981, 78:3824-3828 (T.P.Hopp & K.R.Woods).

The side-chain mass for each of the 20 amino acids.

CRC Handbook of Chemistry and Physics, 66th ed., CRC Press, Boca Raton, Florida (1985).

R.M.C. Dawson, D.C. Elliott, W.H. Elliott, K.M. Jones, Data for Biochemical Research 3rd ed., 

Clarendon Press Oxford (1986).

Altered and converted to python 3.6 by Ana Marta Sequeira 05/2019


#########################################################################################
"""

import math

# import scipy


AALetter = ["A", "R", "N", "D", "C", "E", "Q", "G", "H", "I", "L", "K", "M", "F", "P", "S", "T", "W", "Y", "V"]

_Hydrophobicity = {"A": 0.62, "R": -2.53, "N": -0.78, "D": -0.90, "C": 0.29, "Q": -0.85, "E": -0.74, "G": 0.48,
                   "H": -0.40, "I": 1.38, "L": 1.06, "K": -1.50, "M": 0.64, "F": 1.19, "P": 0.12, "S": -0.18,
                   "T": -0.05, "W": 0.81, "Y": 0.26, "V": 1.08}

_hydrophilicity = {"A": -0.5, "R": 3.0, "N": 0.2, "D": 3.0, "C": -1.0, "Q": 0.2, "E": 3.0, "G": 0.0, "H": -0.5,
                   "I": -1.8, "L": -1.8, "K": 3.0, "M": -1.3, "F": -2.5, "P": 0.0, "S": 0.3, "T": -0.4, "W": -3.4,
                   "Y": -2.3, "V": -1.5}

_residuemass = {"A": 15.0, "R": 101.0, "N": 58.0, "D": 59.0, "C": 47.0, "Q": 72.0, "E": 73.0, "G": 1.000, "H": 82.0,
                "I": 57.0, "L": 57.0, "K": 73.0, "M": 75.0, "F": 91.0, "P": 42.0, "S": 31.0, "T": 45.0, "W": 130.0,
                "Y": 107.0, "V": 43.0}

_pK1 = {"A": 2.35, "C": 1.71, "D": 1.88, "E": 2.19, "F": 2.58, "G": 2.34, "H": 1.78, "I": 2.32, "K": 2.20, "L": 2.36,
        "M": 2.28, "N": 2.18, "P": 1.99, "Q": 2.17, "R": 2.18, "S": 2.21, "T": 2.15, "V": 2.29, "W": 2.38, "Y": 2.20}

_pK2 = {"A": 9.87, "C": 10.78, "D": 9.60, "E": 9.67, "F": 9.24, "G": 9.60, "H": 8.97, "I": 9.76, "K": 8.90, "L": 9.60,
        "M": 9.21, "N": 9.09, "P": 10.6, "Q": 9.13, "R": 9.09, "S": 9.15, "T": 9.12, "V": 9.74, "W": 9.39, "Y": 9.11}

_pI = {"A": 6.11, "C": 5.02, "D": 2.98, "E": 3.08, "F": 5.91, "G": 6.06, "H": 7.64, "I": 6.04, "K": 9.47, "L": 6.04,
       "M": 5.74, "N": 10.76, "P": 6.30, "Q": 5.65, "R": 10.76, "S": 5.68, "T": 5.60, "V": 6.02, "W": 5.88, "Y": 5.63}


#############################################################################################


def _mean(listvalue):
    """
	########################################################################################
	The mean value of the list data.

	result=_mean(listvalue)
	########################################################################################
	"""
    return sum(listvalue) / len(listvalue)


##############################################################################################
def _std(listvalue, ddof=1):
    """
	########################################################################################
	The standard deviation of the list data.

	result=_std(listvalue)
	########################################################################################
	"""
    mean = _mean(listvalue)
    temp = [math.pow(i - mean, 2) for i in listvalue]
    res = math.sqrt(sum(temp) / (len(listvalue) - ddof))
    return res


##############################################################################################


def normalize_each_aap(aap):
    """
	All of the amino acid indices are centralized and standardized before the calculation.

	:param aap: dict form containing the properties of 20 amino acids
	:return: dict form containing the normalized properties of 20 amino acids.
	"""
    if len(list(aap.values())) != 20:
        raise Exception('You can not input the correct number of properties of Amino acids!')
    else:
        Result = {}
        for i, j in list(aap.items()):
            Result[i] = (j - _mean(list(aap.values()))) / _std(list(aap.values()), ddof=0)

        return Result


#############################################################################################
#############################################################################################
##################################Type I pydpi_py3#########################################
####################### Pseudo-Amino Acid Composition ############################
#############################################################################################
#############################################################################################


def _get_correlation_function(Ri='S', Rj='D', aap=None):
    """
	Computing the correlation between two given amino acids using the above three properties.

	:param Ri: amino acids
	:param Rj: amino acids
	:param aap:
	:return: correlation value between two amino acids.
	"""
    if aap is None:
        aap = (_Hydrophobicity, _hydrophilicity, _residuemass)
    hydrophobicity = normalize_each_aap(aap[0])
    hydrophilicity = normalize_each_aap(aap[1])
    residuemass = normalize_each_aap(aap[2])
    theta1 = math.pow(hydrophobicity[Ri] - hydrophobicity[Rj], 2)
    theta2 = math.pow(hydrophilicity[Ri] - hydrophilicity[Rj], 2)
    theta3 = math.pow(residuemass[Ri] - residuemass[Rj], 2)
    theta = round((theta1 + theta2 + theta3) / 3.0, 3)
    return theta


def _get_sequence_order_correlation_factor(protein_sequence, k=1):
    """
	Computing the Sequence order correlation factor with gap equal to k based on [_Hydrophobicity,_hydrophilicity,_residuemass]

	:param protein_sequence: protein is a pure protein sequence
	:param k: gap
	:return: correlation factor value with the gap equal to k
	"""
    length_sequence = len(protein_sequence)
    res = []
    for i in range(length_sequence - k):
        aa1 = protein_sequence[i]
        aa2 = protein_sequence[i + k]
        res.append(_get_correlation_function(aa1, aa2))
    result = round(sum(res) / (length_sequence - k), 3)
    return result


def get_aa_composition(protein_sequence):
    """
	Calculate the composition of Amino acids for a given protein sequence.

	:param protein_sequence: protein is a pure protein sequence
	:return: dict form containing the composition of 20 amino acids.
	"""
    length_sequence = len(protein_sequence)
    result = {}
    for i in AALetter:
        result[i] = round(float(protein_sequence.count(i)) / length_sequence * 100, 3)
    return result


def _get_pseudo_aac1(protein_sequence, lamda=10, weight=0.05):
    """
	Computing the first 20 of type I pseudo-amino acid compostion based on [_Hydrophobicity,_hydrophilicity,_residuemass].

	:param protein_sequence:
	:param lamda:
	:param weight:
	:return:
	"""
    rightpart = 0.0
    for i in range(lamda):
        rightpart = rightpart + _get_sequence_order_correlation_factor(protein_sequence, k=i + 1)
    aac = get_aa_composition(protein_sequence)

    result = {}
    temp = 1 + weight * rightpart
    for index, i in enumerate(AALetter):
        result['PAAC' + str(index + 1)] = round(aac[i] / temp, 3)

    return result


def _get_pseudo_aac2(protein_sequence, lamda=10, weight=0.05):
    """
	Computing the last lamda of type I pseudo-amino acid compostion based on [_Hydrophobicity,_hydrophilicity,_residuemass].

	:param protein_sequence:
	:param lamda:
	:param weight:
	:return:
	"""
    rightpart = []
    for i in range(lamda):
        rightpart.append(_get_sequence_order_correlation_factor(protein_sequence, k=i + 1))

    result = {}
    temp = 1 + weight * sum(rightpart)
    for index in range(20, 20 + lamda):
        result['PAAC' + str(index + 1)] = round(weight * rightpart[index - 20] / temp * 100, 3)

    return result


def _get_pseudo_aac(ProteinSequence, lamda=10, weight=0.05):
    """
    Computing all of type I pseudo-amino acid compostion based on three given properties.
    Note that the number of PAAC strongly depends on the lamda value. if lamda = 20, we can obtain 20+20=40 PAAC.
    The size of these values depends on the	choice of lamda and weight simultaneously.
    AAP=[_Hydrophobicity,_hydrophilicity,_residuemass]

    :param ProteinSequence: protein is a pure protein sequence.
    :param lamda: lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15
        Note that (1)lamda should NOT be larger than the length of input protein sequence;
        (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the
        output of PseAA server is the 20-D amino acid composition.
    :param weight:
        weight factor is designed for the users to put weight on the additional PseAA components. with respect to the
        conventional AA components. The user can select any value within the region from 0.05
        to 0.7 for the weight factor.

    :return: dict form containing calculated 20+lamda PAAC
    """
    res = {}
    res.update(_get_pseudo_aac1(ProteinSequence, lamda=lamda, weight=weight))
    res.update(_get_pseudo_aac2(ProteinSequence, lamda=lamda, weight=weight))
    return res


#############################################################################################
##################################Type II pydpi_py3########################################
###############Amphiphilic Pseudo-Amino Acid Composition pydpi_py3#########################
#############################################################################################
#############################################################################################


def _get_correlation_function_for_apaac(Ri='S', Rj='D', AAP=None):
    """
	Computing the correlation between two given amino acids using the above two	properties for APAAC (type II PseAAC).

	:param Ri: amino acids
	:param Rj: amino acids
	:param AAP:
	:return:  correlation value between two amino acids
	"""
    if AAP is None:
        AAP= (_Hydrophobicity, _hydrophilicity)
    hydrophobicity = normalize_each_aap(AAP[0])
    hydrophilicity = normalize_each_aap(AAP[1])
    theta1 = round(hydrophobicity[Ri] * hydrophobicity[Rj], 3)
    theta2 = round(hydrophilicity[Ri] * hydrophilicity[Rj], 3)

    return theta1, theta2


def get_sequence_order_correlation_factor_for_apaac(protein_sequence, k=1):
    """
	Computing the Sequence order correlation factor with gap equal to k based on [_Hydrophobicity,_hydrophilicity]
	for APAAC (type II PseAAC) .

	:param protein_sequence: protein is a pure protein sequence
	:param k: gap
	:return: correlation factor value with the gap equal to k
	"""
    length_sequence = len(protein_sequence)
    res_hydrophobicity = []
    reshydrophilicity = []
    for i in range(length_sequence - k):
        AA1 = protein_sequence[i]
        AA2 = protein_sequence[i + k]
        temp = _get_correlation_function_for_apaac(AA1, AA2)
        res_hydrophobicity.append(temp[0])
        reshydrophilicity.append(temp[1])
    result = []
    result.append(round(sum(res_hydrophobicity) / (length_sequence - k), 3))
    result.append(round(sum(reshydrophilicity) / (length_sequence - k), 3))
    return result


def get_a_pseudo_aac1(protein_sequence, lamda=30, weight=0.5):
    """
    Computing the first 20 of type II pseudo-amino acid compostion based on	[_Hydrophobicity,_hydrophilicity].

    :param protein_sequence:
    :param lamda:
    :param weight:
    :return:
    """
    rightpart = 0.0
    for i in range(lamda):
        rightpart = rightpart + sum(get_sequence_order_correlation_factor_for_apaac(protein_sequence, k=i + 1))
    AAC = get_aa_composition(protein_sequence)

    result = {}
    temp = 1 + weight * rightpart
    for index, i in enumerate(AALetter):
        result['APAAC' + str(index + 1)] = round(AAC[i] / temp, 3)

    return result


def get_a_pseudo_aac2(protein_sequence, lamda=30, weight=0.5):
    """
	Computing the last lamda of type II pseudo-amino acid compostion based on	[_Hydrophobicity,_hydrophilicity].

	:param protein_sequence:
	:param lamda:
	:param weight:
	:return:
	"""
    rightpart = []
    for i in range(lamda):
        temp = get_sequence_order_correlation_factor_for_apaac(protein_sequence, k=i + 1)
        rightpart.append(temp[0])
        rightpart.append(temp[1])

    result = {}
    temp = 1 + weight * sum(rightpart)
    for index in range(20, 20 + 2 * lamda):
        result['APAAC' + str(index + 1)] = round(weight * rightpart[index - 20] / temp * 100, 3)

    return result


def get_a_pseudo_aac(protein_sequence, lamda=30, weight=0.5):
    """
    Computing all of type II pseudo-amino acid compostion based on the given
    properties. Note that the number of PAAC strongly depends on the lamda value. if lamda = 20,
    we can obtain 20+20=40 PAAC. The size of these values depends on the choice of lamda and weight simultaneously.

    :param protein_sequence: protein is a pure protein sequence
    :param lamda: lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.
        Note that (1)lamda should NOT be larger than the length of input protein sequence;
        (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the
        output of PseAA server is the 20-D amino acid composition.
    :param weight: weight factor is designed for the users to put weight on the additional PseAA components
        with respect to the conventional AA components. The user can select any value within the
        region from 0.05 to 0.7 for the weight factor.
    :return: dict form containing calculated 20+lamda PAAC

    """
    res = {}
    res.update(get_a_pseudo_aac1(protein_sequence, lamda=lamda, weight=weight))
    res.update(get_a_pseudo_aac2(protein_sequence, lamda=lamda, weight=weight))
    return res


#############################################################################################
#############################################################################################
##################################Type I pydpi_py3#########################################
####################### Pseudo-Amino Acid Composition pydpi_py3############################
#############################based on different properties###################################
#############################################################################################
#############################################################################################
def get_correlation_function(Ri='S', Rj='D', AAP= None):
    """
	Computing the correlation between two given amino acids using the given properties.

	:param Ri: amino acids
	:param Rj: amino acids
	:param AAP: list form containing the properties, each of which is a dict form
	:return: correlation value between two amino acids
	"""
    if AAP is None:
        AAP = [_Hydrophobicity, _hydrophilicity, _residuemass, _pK1, _pK2, _pI]
    NumAAP = len(AAP)
    theta = 0.0
    for i in range(NumAAP):
        temp = normalize_each_aap(AAP[i])
        theta = theta + math.pow(temp[Ri] - temp[Rj], 2)
    result = round(theta / NumAAP, 3)
    return result


def get_sequence_order_correlation_factor(protein_sequence, k=1, AAP=None):
    """
	Computing the Sequence order correlation factor with gap equal to k based on the given properities.

	:param protein_sequence: protein is a pure protein sequence
	:param k: gap
	:param AAP: AAP is a list form containing the properties, each of which is a dict form
	:return: correlation factor value with the gap equal to k
	"""

    length_sequence = len(protein_sequence)
    res = []
    for i in range(length_sequence - k):
        aa1 = protein_sequence[i]
        aa2 = protein_sequence[i + k]
        res.append(get_correlation_function(aa1, aa2, AAP))
    result = round(sum(res) / (length_sequence - k), 3)
    return result


def get_pseudo_aac1(protein_sequence, lamda=30, weight=0.05, AAP=None):
    """
	Computing the first 20 of type I pseudo-amino acid compostion based on the given properties.

	:param protein_sequence:
	:param lamda:
	:param weight:
	:param AAP:
	:return:
	"""
    rightpart = 0.0
    for i in range(lamda):
        rightpart = rightpart + get_sequence_order_correlation_factor(protein_sequence, i + 1, AAP)
    aac = get_aa_composition(protein_sequence)

    result = {}
    temp = 1 + weight * rightpart
    for index, i in enumerate(AALetter):
        result['PAAC' + str(index + 1)] = round(aac[i] / temp, 3)

    return result


def get_pseudo_aac2(protein_sequence, lamda=30, weight=0.05, AAP=None):
    """
	Computing the last lamda of type I pseudo-amino acid compostion based on the given properties.

	:param protein_sequence:
	:param lamda:
	:param weight:
	:param AAP:
	:return:
	"""
    rightpart = []
    for i in range(lamda):
        rightpart.append(get_sequence_order_correlation_factor(protein_sequence, i + 1, AAP))

    result = {}
    temp = 1 + weight * sum(rightpart)
    for index in range(20, 20 + lamda):
        result['PAAC' + str(index + 1)] = round(weight * rightpart[index - 20] / temp * 100, 3)

    return result


def get_pseudo_aac(protein_sequence, lamda=30, weight=0.05, AAP=None):
    """
    Computing all of type I pseudo-amino acid compostion based on the given
    properties. Note that the number of PAAC strongly depends on the lamda value. if lamda = 20,
    we can obtain 20+20=40 PAAC. The size of these values depends on the choice of lamda and weight simultaneously.
    You must specify some properties into AAP.

    :param protein_sequence: protein is a pure protein sequence
    :param lamda: 	lamda factor reflects the rank of correlation and is a non-Negative integer, such as 15.
        Note that (1)lamda should NOT be larger than the length of input protein sequence;
        (2) lamda must be non-Negative integer, such as 0, 1, 2, ...; (3) when lamda =0, the
        output of PseAA server is the 20-D amino acid composition.
    :param weight: weight factor is designed for the users to put weight on the additional PseAA components
        with respect to the conventional AA components. The user can select any value within the
        region from 0.05 to 0.7 for the weight factor.
    :param AAP: AAP is a list form containing the properties, each of which is a dict form

    :return: dict form containing calculated 20+lamda PAAC
    """
    res = {}
    res.update(get_pseudo_aac1(protein_sequence, lamda, weight, AAP))
    res.update(get_pseudo_aac2(protein_sequence, lamda, weight, AAP))
    return res


if __name__ == "__main__":

    protein = "MTDRARLRLHDTAAGVVRDFVPLRPGHVSIYLCGATVQGLPHIGHVRSGVAFDILRRWLL\
ARGYDVAFIRNVTDIEDKILAKAAAAGRPWWEWAATHERAFTAAYDALDVLPPSAEPRAT\
GHITQMIEMIERLIQAGHAYTGGGDVYFDVLSYPEYGQLSGHKIDDVHQGEGVAAGKRDQ\
RDFTLWKGEKPGEPSWPTPWGRGRPGWHLECSAMARSYLGPEFDIHCGGMDLVFPHHENE\
IAQSRAAGDGFARYWLHNGWVTMGGEKMSKSLGNVLSMPAMLQRVRPAELRYYLGSAHYR\
SMLEFSETAMQDAVKAYVGLEDFLHRVRTRVGAVCPGDPTPRFAEALDDDLSVPIALAEI\
HHVRAEGNRALDAGDHDGALRSASAIRAMMGILGCDPLDQRWESRDETSAALAAVDVLVQ\
AELQNREKAREQRNWALADEIRGRLKRAGIEVTDTADGPQWSLLGGDTK"
    protein = str.strip(protein)
    PAAC = get_pseudo_aac(protein, lamda=5, AAP=None)

    for i in PAAC:
        print(i, PAAC[i])

    protein = "AAGMGFFGAR"
    protein = str.strip(protein)
    PAAC = get_pseudo_aac(protein, lamda=5, AAP=[])

    for i in PAAC:
        print(i, PAAC[i])




# -*- coding: utf-8 -*-
"""
.. currentmodule:: modlamp.pydpi_py3

.. moduleauthor:: modlab Alex Mueller ETH Zurich <alex.mueller@pharma.ethz.ch>

This module incorporates different classes to calculate peptide descriptor values. The following classes are available:

=============================        ============================================================================
Class                                Characteristics
=============================        ============================================================================
:py:class:`GlobalDescriptor`         Global one-dimensional peptide pydpi_py3 calculated from the AA sequence.
:py:class:`PeptideDescriptor`        AA scale based global or convoluted pydpi_py3 (auto-/cross-correlated).
=============================        ============================================================================

.. seealso:: :class:`modlamp.core.BaseDescriptor` from which the classes in :mod:`modlamp.pydpi_py3` inherit.
"""

import sys

import numpy as np
from scipy import stats
# from sklearn.externals.joblib import Parallel, delayed
from joblib import Parallel, delayed

from .modlamp_core import BaseDescriptor, load_scale, count_aas, aa_weights, aa_energies, aa_formulas

__author__ = "Alex Müller, Gisela Gabernet"
__docformat__ = "restructuredtext en"


def _one_autocorr(seq, window, scale):
    """Private function used for calculating auto-correlated pydpi for 1 given sequence, window and an AA scale.
    This function is used by the :py:func:`calculate_autocorr` method of :py:class:`PeptideDescriptor`.

    :param seq: {str} amino acid sequence to calculate descriptor for
    :param window: {int} correlation-window size
    :param scale: {str} amino acid scale to be used to calculate descriptor
    :return: {numpy.array} calculated descriptor data
    """
    try:
        m = list()  # list of lists to store translated sequence values
        for l in range(len(seq)):  # translate AA sequence into values
            m.append(scale[str(seq[l])])
        # auto-correlation in defined sequence window
        seqdesc = list()
        for dist in range(window):  # for all correlation distances
            for val in range(len(scale['A'])):  # for all features of the descriptor scale
                valsum = list()
                cntr = 0.
                for pos in range(len(seq)):  # for every position in the sequence
                    if (pos + dist) < len(seq):  # check if corr distance is possible at that sequence position
                        cntr += 1  # counter to scale sum
                        valsum.append(m[pos][val] * m[pos + dist][val])
                seqdesc.append(sum(valsum) / cntr)  # append scaled correlation distance values
        return seqdesc
    except ZeroDivisionError:
        print("ERROR!\nThe chosen correlation window % i is larger than the sequence %s !" % (window, seq))
        

def _one_crosscorr(seq, window, scale):
    """Private function used for calculating cross-correlated pydpi_py3 for 1 given sequence, window and an AA scale.
    This function is used by the :py:func:`calculate_crosscorr` method of :py:class:`PeptideDescriptor`.

    :param seq: {str} amino acid sequence to calculate descriptor for
    :param window: {int} correlation-window size
    :param scale: {str} amino acid scale to be used to calculate descriptor
    :return: {numpy.array} calculated descriptor data
    """
    try:
        m = list()  # list of lists to store translated sequence values
        for l in range(len(seq)):  # translate AA sequence into values
            m.append(scale[str(seq[l])])
        # auto-correlation in defined sequence window
        seqdesc = list()
        for val in range(len(scale['A'])):  # for all features of the descriptor scale
            for cc in range(len(scale['A'])):  # for every feature cross correlation
                if (val + cc) < len(scale['A']):  # check if corr distance is in range of the num of features
                    for dist in range(window):  # for all correlation distances
                        cntr = float()
                        valsum = list()
                        for pos in range(len(seq)):  # for every position in the sequence
                            if (pos + dist) < len(seq):  # check if corr distance is possible at that sequence pos
                                cntr += 1  # counter to scale sum
                                valsum.append(m[pos][val] * m[pos + dist][val + cc])
                        seqdesc.append(sum(valsum) / cntr)  # append scaled correlation distance values
        return seqdesc
    except ZeroDivisionError:
        print("ERROR!\nThe chosen correlation window % i is larger than the sequence %s !" % (window, seq))


def _one_arc(seq, modality, scale):
    """ Privat function used for calculating arc pydpi_py3 for one sequence and AA scale. This function is used by
    :py:func:`calculate_arc` method method of :py:class:`PeptideDescriptor`.

    :param seq: {str} amino acid sequence to calculate descriptor for
    :param scale: {str} amino acid scale to be used to calculate descriptor
    :return: {numpy.array} calculated descriptor data
    """
    desc_mat = []
    for aa in seq:
        desc_mat.append(scale[aa])
    desc_mat = np.asarray(desc_mat)

    # Check descriptor dimension
    desc_dim = desc_mat.shape[1]

    # list to store descriptor values for all windows
    allwindows_arc = []

    if len(seq) > 18:
        window = 18
        # calculates number of windows in sequence
        num_windows = len(seq) - window
    else:
        window = len(seq)
        num_windows = 1

    # loop through all windows
    for j in range(num_windows):
        # slices descriptor matrix into current window
        window_mat = desc_mat[j:j + window, :]

        # defines order of amino acids in helical projection
        order = [0, 11, 4, 15, 8, 1, 12, 5, 16, 9, 2, 13, 6, 17, 10, 3, 14, 7]

        # orders window descriptor matrix into helical projection order
        ordered = []
        for pos in order:
            try:
                ordered.append(window_mat[pos, :])
            except:
                # for sequences of len < 18 adding dummy vector with 2s, length of descriptor dimensions
                ordered.append([2] * desc_dim)
        ordered = np.asarray(ordered)

        window_arc = []

        # loop through pharmacophoric features
        for m in range(desc_dim):
            all_arcs = []  # stores all arcs that can be found of a pharmacophoric feature
            arc = 0

            for n in range(18):  # for all positions in helix, regardless of sequence length
                if ordered[n, m] == 0:  # if position does not contain pharmacophoric feature
                    all_arcs.append(arc)  # append previous arc to all arcs list
                    arc = 0  # arc is initialized
                elif ordered[n, m] == 1:  # if position contains pharmacophoric feature(PF), elongate arc by 20°
                    arc += 20
                elif ordered[n, m] == 2:  # if position doesn't contain amino acid:
                    if ordered[n - 1, m] == 1:  # if previous position contained PF add 10°
                        arc += 10
                    elif ordered[n - 1, m] == 0:  # if previous position didn't contain PF don't add anything
                        arc += 0
                    elif ordered[
                                n - 2, m] == 1:  # if previous position is empty then check second previous for PF
                        arc += 10
                    if n == 17:  # if we are at the last position check for position n=0 instead of next position.
                        if ordered[0, m] == 1:  # if it contains PF add 10° extra
                            arc += 10
                    else:  # if next position contains PF add 10° extra
                        if ordered[n + 1, m] == 1:
                            arc += 10
                        elif ordered[n + 1, m] == 0:
                            arc += 0
                        else:  # if next position is empty check for 2nd next position
                            if n == 16:
                                if ordered[0, m] == 1:
                                    arc += 10
                            else:
                                if ordered[n + 2, m] == 1:
                                    arc += 10

            all_arcs.append(arc)
            if not arc == 360:
                arc0 = all_arcs.pop() + all_arcs[0]  # join first and last arc together
                all_arcs = [arc0] + all_arcs[1:]

            window_arc.append(np.max(all_arcs))  # append to window arcs the maximum arc of this PF
        allwindows_arc.append(window_arc)  # append all PF arcs of this window

    allwindows_arc = np.asarray(allwindows_arc)

    if modality == 'max':
        final_arc = np.max(allwindows_arc, axis=0)  # calculate maximum / mean arc along all windows
    elif modality == 'mean':
        final_arc = np.mean(allwindows_arc, axis=0)
    else:
        print('modality is unknown, please choose between "max" and "mean"\n.')
        sys.exit()
    return final_arc


def _charge(seq, ph=7.0, amide=False):
    """Calculates charge of a single sequence. The method used is first described by Bjellqvist. In the case of
    amidation, the value for the  'Cterm' pKa is 15 (and Cterm is added to the pos_pks dictionary.
    The pKa scale is extracted from: http://www.hbcpnetbase.com/ (CRC Handbook of Chemistry and Physics, 96th ed).

    **pos_pks** = {'Nterm': 9.38, 'K': 10.67, 'R': 12.10, 'H': 6.04}

    **neg_pks** = {'Cterm': 2.15, 'D': 3.71, 'E': 4.15, 'C': 8.14, 'Y': 10.10}

    :param ph: {float} pH at which to calculate peptide charge.
    :param amide: {boolean} whether the sequences have an amidated C-terminus.
    :return: {array} descriptor values in the attribute :py:attr:`descriptor
    """
    
    if amide:
        pos_pks = {'Nterm': 9.38, 'K': 10.67, 'R': 12.10, 'H': 6.04}
        neg_pks = {'Cterm': 15., 'D': 3.71, 'E': 4.15, 'C': 8.14, 'Y': 10.10}
    else:
        pos_pks = {'Nterm': 9.38, 'K': 10.67, 'R': 12.10, 'H': 6.04}
        neg_pks = {'Cterm': 2.15, 'D': 3.71, 'E': 4.15, 'C': 8.14, 'Y': 10.10}
    
    aa_content = count_aas(seq, scale='absolute')
    aa_content['Nterm'] = 1.0
    aa_content['Cterm'] = 1.0
    pos_charge = 0.0
    for aa, pK in pos_pks.items():
        c_r = 10 ** (pK - ph)
        partial_charge = c_r / (c_r + 1.0)
        pos_charge += aa_content[aa] * partial_charge
    neg_charge = 0.0
    for aa, pK in neg_pks.items():
        c_r = 10 ** (ph - pK)
        partial_charge = c_r / (c_r + 1.0)
        neg_charge += aa_content[aa] * partial_charge
    return round(pos_charge - neg_charge, 3)


class GlobalDescriptor(BaseDescriptor):
    """
    Base class for global, non-amino acid scale dependant pydpi_py3. The following pydpi_py3 can be calculated by
    the **methods** linked below:

    - `Sequence Length      <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.length>`_
    - `Molecular Formula    <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.formula>`_
    - `Molecular Weight     <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.calculate_MW>`_
    - `Sequence Charge      <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.calculate_charge>`_
    - `Charge Density       <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.charge_density>`_
    - `Isoelectric Point    <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.isoelectric_point>`_
    - `Instability Index    <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.instability_index>`_
    - `Aromaticity          <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.aromaticity>`_
    - `Aliphatic Index      <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.aliphatic_index>`_
    - `Boman Index          <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.boman_index>`_
    - `Hydrophobic Ratio    <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.hydrophobic_ratio>`_
    - `all of the above     <modlamp.html#modlamp.pydpi_py3.GlobalDescriptor.calculate_all>`_
    """

    def length(self, append=False):
        """
        Method to calculate the length (total AA count) of every sequence in the attribute :py:attr:`sequences`.

        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of sequence lengths in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor(['AFDGHLKI','KKLQRSDLLRTK','KKLASCNNIPPR'])
        >>> desc.length()
        >>> desc.descriptor
        array([[ 8.], [12.], [12.]])
        """
        desc = []
        for seq in self.sequences:
            desc.append(float(len(seq.strip())))
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('Length')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['Length']

    def formula(self, amide=False, append=False):
        """Method to calculate the molecular formula of every sequence in the attribute :py:attr:`sequences`.
        
        :param amide: {boolean} whether the sequences are C-terminally amidated.
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of molecular formulas {str} in the attribute :py:attr:`descriptor`
        :Example:
        
        # >>> desc = GlobalDescriptor(['KADSFLSADGHSADFSLDKKLKERL', 'ERTILSDFPQWWFASLDFLNC', 'ACDEFGHIKLMNPQRSTVWY'])
        # >>> desc.formula(amide=True)
        # >>> for v in desc.descriptor:
        # ...     print(v[0])
        C122 H197 N35 O39
        C121 H168 N28 O33 S
        C106 H157 N29 O30 S2
        
        .. seealso:: :py:func:`modlamp.core.aa_formulas()`
        
        .. versionadded:: v2.7.6
        """
        desc = []
        formulas = aa_formulas()
        for seq in self.sequences:
            f = {'C': 0, 'H': 0, 'N': 0, 'O': 0, 'S': 0}
            for aa in seq:  # sum over aa weights
                for k in f.keys():
                    f[k] += formulas[aa][k]
            
            # substract H2O for every peptide bond
            f['H'] -= 2 * (len(seq) - 1)
            f['O'] -= (len(seq) - 1)
            
            if amide:  # add C-terminal amide --> replace OH with NH2
                f['O'] -= 1
                f['H'] += 1
                f['N'] += 1
            
            if f['S'] != 0:
                val = 'C%s H%s N%s O%s %s%s' % (f['C'], f['H'], f['N'], f['O'], 'S', f['S'])
            else:
                val = 'C%s H%s N%s O%s' % (f['C'], f['H'], f['N'], f['O'])
                
            desc.append([val])
        
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('Formula')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['Formula']

    def calculate_MW(self, amide=False, append=False):
        """Method to calculate the molecular weight [g/mol] of every sequence in the attribute :py:attr:`sequences`.

        :param amide: {boolean} whether the sequences are C-terminally amidated (subtracts 0.95 from the MW).
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('IAESFKGHIPL')
        >>> desc.calculate_MW(amide=True)
        >>> desc.descriptor
        array([[ 1210.43]])

        .. seealso:: :py:func:`modlamp.core.aa_weights()`

        .. versionchanged:: v2.1.5 amide option added
        """
        desc = []
        weights = aa_weights()
        for seq in self.sequences:
            mw = []
            for aa in seq:  # sum over aa weights
                mw.append(weights[aa])
            desc.append(round(sum(mw) - 18.015 * (len(seq) - 1), 2))  # sum over AA MW and subtract H20 MW for every
            # peptide bond
        desc = np.asarray(desc).reshape(len(desc), 1)
        if amide:  # if sequences are amidated, subtract 0.98 from calculated MW (OH - NH2)
            desc = [d - 0.98 for d in desc]
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('MW')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['MW']

    def calculate_charge(self, ph=7.0, amide=False, append=False):
        """Method to overall charge of every sequence in the attribute :py:attr:`sequences`.

        The method used is first described by Bjellqvist. In the case of amidation, the value for the 'Cterm' pKa is 15
        (and Cterm is added to the pos_pKs dictionary.
        The pKa scale is extracted from: http://www.hbcpnetbase.com/ (CRC Handbook of Chemistry and Physics, 96th ed).

        **pos_pKs** = {'Nterm': 9.38, 'K': 10.67, 'R': 12.10, 'H': 6.04}

        **neg_pKs** = {'Cterm': 2.15, 'D': 3.71, 'E': 4.15, 'C': 8.14, 'Y': 10.10}

        :param ph: {float} ph at which to calculate peptide charge.
        :param amide: {boolean} whether the sequences have an amidated C-terminus.
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('KLAKFGKRSELVALSG')
        >>> desc.calculate_charge(ph=7.4, amide=True)
        >>> desc.descriptor
        array([[ 3.989]])
        """

        desc = []
        for seq in self.sequences:
            desc.append(_charge(seq, ph, amide))  # calculate charge with helper function
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('Charge')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['Charge']

    def charge_density(self, ph=7.0, amide=False, append=False):
        """Method to calculate the charge density (charge / MW) of every sequences in the attributes :py:attr:`sequences`

        :param ph: {float} pH at which to calculate peptide charge.
        :param amide: {boolean} whether the sequences have an amidated C-terminus.
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`.
        :Example:
        
        >>> desc = GlobalDescriptor('GNSDLLIEQRTLLASDEF')
        >>> desc.charge_density(ph=6, amide=True)
        >>> desc.descriptor
        array([[-0.00097119]])
        """
        self.calculate_charge(ph, amide)
        charges = self.descriptor
        self.calculate_MW(amide)
        masses = self.descriptor
        desc = charges / masses
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('ChargeDensity')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['ChargeDensity']

    def isoelectric_point(self, amide=False, append=False):
        """
        Method to calculate the isoelectric point of every sequence in the attribute :py:attr:`sequences`.
        The pK scale is extracted from: http://www.hbcpnetbase.com/ (CRC Handbook of Chemistry and Physics, 96th ed).

         **pos_pKs** = {'Nterm': 9.38, 'K': 10.67, 'R': 12.10, 'H': 6.04}

         **neg_pKs** = {'Cterm': 2.15, 'D': 3.71, 'E': 4.15, 'C': 8.14, 'Y': 10.10}

        :param amide: {boolean} whether the sequences have an amidated C-terminus.
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        # >>> desc = GlobalDescriptor('KLFDIKFGHIPQRST')
        # >>> desc.isoelectric_point()
        # >>> desc.descriptor
        array([[ 10.6796875]])
        """

        desc = []
        for seq in self.sequences:

            # Bracket between ph1 and ph2
            ph = 7.0
            charge = _charge(seq, ph, amide)
            if charge > 0.0:
                ph1 = ph
                charge1 = charge
                while charge1 > 0.0:
                    ph = ph1 + 1.0
                    charge = _charge(seq, ph, amide)
                    if charge > 0.0:
                        ph1 = ph
                        charge1 = charge
                    else:
                        ph2 = ph
                        break
            else:
                ph2 = ph
                charge2 = charge
                while charge2 < 0.0:
                    ph = ph2 - 1.0
                    charge = _charge(seq, ph, amide)
                    if charge < 0.0:
                        ph2 = ph
                        charge2 = charge
                    else:
                        ph1 = ph
                        break
            # Bisection
            while ph2 - ph1 > 0.0001 and charge != 0.0:
                ph = (ph1 + ph2) / 2.0
                charge = _charge(seq, ph, amide)
                if charge > 0.0:
                    ph1 = ph
                else:
                    ph2 = ph
            desc.append(ph)
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('pI')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['pI']

    def instability_index(self, append=False):
        """
        Method to calculate the instability of every sequence in the attribute :py:attr:`sequences`.
        The instability index is a prediction of protein stability based on the amino acid composition.
        ([1] K. Guruprasad, B. V Reddy, M. W. Pandit, Protein Eng. 1990, 4, 155–161.)

        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('LLASMNDLLAKRST')
        >>> desc.instability_index()
        >>> desc.descriptor
        array([[ 63.95714286]])
        """
        
        desc = []
        dimv = load_scale('instability')[1]
        for seq in self.sequences:
            stabindex = float()
            for i in range(len(seq) - 1):
                stabindex += dimv[seq[i]][seq[i+1]]
            desc.append((10.0 / len(seq)) * stabindex)
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('InstabilityInd')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['InstabilityInd']

    def aromaticity(self, append=False):
        """
        Method to calculate the aromaticity of every sequence in the attribute :py:attr:`sequences`.
        According to Lobry, 1994, it is simply the relative frequency of Phe+Trp+Tyr.

        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('GLFYWRFFLQRRFLYWW')
        >>> desc.aromaticity()
        >>> desc.descriptor
        array([[ 0.52941176]])
        """
        desc = []
        for seq in self.sequences:
            f = seq.count('F')
            w = seq.count('W')
            y = seq.count('Y')
            desc.append(float(f + w + y) / len(seq))
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('Aromaticity')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['Aromaticity']

    def aliphatic_index(self, append=False):
        """
        Method to calculate the aliphatic index of every sequence in the attribute :py:attr:`sequences`.
        According to Ikai, 1980, the aliphatic index is a measure of thermal stability of proteins and is dependant
        on the relative volume occupied by aliphatic amino acids (A,I,L & V).
        ([1] A. Ikai, J. Biochem. 1980, 88, 1895–1898.)

        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('KWLKYLKKLAKLVK')
        >>> desc.aliphatic_index()
        >>> desc.descriptor
        array([[ 139.28571429]])
        """
        desc = []
        aa_dict = aa_weights()
        for seq in self.sequences:
            d = {aa: seq.count(aa) for aa in aa_dict.keys()}  # count aa
            d = {k: (float(d[k]) / len(seq)) * 100 for k in d.keys()}  # get mole percent of all AA
            desc.append(d['A'] + 2.9 * d['V'] + 3.9 * (d['I'] + d['L']))  # formula for calculating the AI (Ikai, 1980)
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('AliphaticInd')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['AliphaticInd']

    def boman_index(self, append=False):
        """Method to calculate the boman index of every sequence in the attribute :py:attr:`sequences`.
        According to Boman, 2003, the boman index is a measure for protein-protein interactions and is calculated by
        summing over all amino acid free energy of transfer [kcal/mol] between water and cyclohexane,[2] followed by
        dividing by    sequence length.
        ([1] H. G. Boman, D. Wade, I. a Boman, B. Wåhlin, R. B. Merrifield, *FEBS Lett*. **1989**, *259*, 103–106.
        [2] A. Radzick, R. Wolfenden, *Biochemistry* **1988**, *27*, 1664–1670.)
        
        .. seealso:: :py:func:`modlamp.core.aa_energies()`

        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('GLFDIVKKVVGALGSL')
        >>> desc.boman_index()
        >>> desc.descriptor
        array([[-1.011875]])
        """
        d = aa_energies()
        desc = []
        for seq in self.sequences:
            val = []
            for a in seq:
                val.append(d[a])
            desc.append(sum(val) / len(val))
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('BomanInd')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['BomanInd']

    def hydrophobic_ratio(self, append=False):
        """
        Method to calculate the hydrophobic ratio of every sequence in the attribute :py:attr:`sequences`, which is the
        relative frequency of the amino acids **A,C,F,I,L,M & V**.

        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('VALLYWRTVLLAIII')
        >>> desc.hydrophobic_ratio()
        >>> desc.descriptor
        array([[ 0.73333333]])
        """
        desc = []
        aa_dict = aa_weights()
        for seq in self.sequences:
            pa = {aa: seq.count(aa) for aa in aa_dict.keys()}  # count aa
            # formula for calculating the AI (Ikai, 1980):
            desc.append((pa['A'] + pa['C'] + pa['F'] + pa['I'] + pa['L'] + pa['M'] + pa['V']) / float(len(seq)))
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            self.featurenames.append('HydrophRatio')
        else:
            self.descriptor = np.array(desc)
            self.featurenames = ['HydrophRatio']

    def calculate_all(self, ph=7.4, amide=True):
        """Method combining all global pydpi_py3 and appending them into the feature matrix in the attribute
        :py:attr:`descriptor`.
        
        :param ph: {float} pH at which to calculate peptide charge
        :param amide: {boolean} whether the sequences have an amidated C-terminus.
        :return: array of descriptor values in the attribute :py:attr:`descriptor`
        :Example:
        
        >>> desc = GlobalDescriptor('AFGHFKLKKLFIFGHERT')
        >>> desc.calculate_all(amide=True)
        >>> desc.featurenames
        ['Length', 'MW', 'ChargeDensity', 'pI', 'InstabilityInd', 'Aromaticity', 'AliphaticInd', 'BomanInd', 'HydRatio']
        >>> desc.descriptor
        array([[ 18.,  2.17559000e+03,   1.87167619e-03,   1.16757812e+01, ...  1.10555556e+00,   4.44444444e-01]])
        >>> desc.save_descriptor('/path/to/outputfile.csv')  # save the descriptor data (with feature names header)
        """
        
        # This is a strange way of doing it. However, the append=True option excludes length and charge, no idea why!
        fn = []
        self.length()  # sequence length
        l = self.descriptor
        fn.extend(self.featurenames)
        self.calculate_MW(amide=amide)  # molecular weight
        mw = self.descriptor
        fn.extend(self.featurenames)
        self.calculate_charge(ph=ph, amide=amide)  # net charge
        c = self.descriptor
        fn.extend(self.featurenames)
        self.charge_density(ph=ph, amide=amide)  # charge density
        cd = self.descriptor
        fn.extend(self.featurenames)
        self.isoelectric_point(amide=amide)  # pI
        pi = self.descriptor
        fn.extend(self.featurenames)
        self.instability_index()  # instability index
        si = self.descriptor
        fn.extend(self.featurenames)
        self.aromaticity()  # global aromaticity
        ar = self.descriptor
        fn.extend(self.featurenames)
        self.aliphatic_index()  # aliphatic index
        ai = self.descriptor
        fn.extend(self.featurenames)
        self.boman_index()  # Boman index
        bi = self.descriptor
        fn.extend(self.featurenames)
        self.hydrophobic_ratio()  # Hydrophobic ratio
        hr = self.descriptor
        fn.extend(self.featurenames)
        
        self.descriptor = np.concatenate((l, mw, c, cd, pi, si, ar, ai, bi, hr), axis=1)
        self.featurenames = fn


class PeptideDescriptor(BaseDescriptor):
    """Base class for peptide pydpi_py3. The following **amino acid descriptor scales** are available for descriptor
    calculation:

    - **AASI**           (An amino acid selectivity index scale for helical antimicrobial peptides, *[1] D. Juretić, D. Vukicević, N. Ilić, N. Antcheva, A. Tossi, J. Chem. Inf. Model. 2009, 49, 2873–2882.*)
    - **ABHPRK**         (modlabs inhouse physicochemical feature scale (Acidic, Basic, Hydrophobic, Polar, aRomatic, Kink-inducer)
    - **argos**          (Argos hydrophobicity amino acid scale, *[2] Argos, P., Rao, J. K. M. & Hargrave, P. A., Eur. J. Biochem. 2005, 128, 565–575.*)
    - **bulkiness**      (Amino acid side chain bulkiness scale, *[3] J. M. Zimmerman, N. Eliezer, R. Simha, J. Theor. Biol. 1968, 21, 170–201.*)
    - **charge_phys**    (Amino acid charge at pH 7.0 - Hystidine charge +0.1.)
    - **charge_acid**    (Amino acid charge at acidic pH - Hystidine charge +1.0.)
    - **cougar**         (modlabs inhouse selection of global peptide pydpi_py3)
    - **eisenberg**      (the Eisenberg hydrophobicity consensus amino acid scale, *[4] D. Eisenberg, R. M. Weiss, T. C. Terwilliger, W. Wilcox, Faraday Symp. Chem. Soc. 1982, 17, 109.*)
    - **Ez**             (potential that assesses energies of insertion of amino acid side chains into lipid bilayers, *[5] A. Senes, D. C. Chadi, P. B. Law, R. F. S. Walters, V. Nanda, W. F. DeGrado, J. Mol. Biol. 2007, 366, 436–448.*)
    - **flexibility**    (amino acid side chain flexibilitiy scale, *[6] R. Bhaskaran, P. K. Ponnuswamy, Int. J. Pept. Protein Res. 1988, 32, 241–255.*)
    - **grantham**       (amino acid side chain composition, polarity and molecular volume, *[8] Grantham, R. Science. 185, 862–864 (1974).*)
    - **gravy**          (GRAVY hydrophobicity amino acid scale, *[9] J. Kyte, R. F. Doolittle, J. Mol. Biol. 1982, 157, 105–132.*)
    - **hopp-woods**     (Hopp-Woods amino acid hydrophobicity scale,*[10] T. P. Hopp, K. R. Woods, Proc. Natl. Acad. Sci. 1981, 78, 3824–3828.*)
    - **ISAECI**         (Isotropic Surface Area (ISA) and Electronic Charge Index (ECI) of amino acid side chains, *[11] E. R. Collantes, W. J. Dunn, J. Med. Chem. 1995, 38, 2705–2713.*)
    - **janin**          (Janin hydrophobicity amino acid scale, *[12] J. L. Cornette, K. B. Cease, H. Margalit, J. L. Spouge, J. A. Berzofsky, C. DeLisi, J. Mol. Biol. 1987, 195, 659–685.*)
    - **kytedoolittle**  (Kyte & Doolittle hydrophobicity amino acid scale, *[13] J. Kyte, R. F. Doolittle, J. Mol. Biol. 1982, 157, 105–132.*)
    - **levitt_alpha**   (Levitt amino acid alpha-helix propensity scale, extracted from http://web.expasy.org/protscale. *[14] M. Levitt, Biochemistry 1978, 17, 4277-4285.*)
    - **MSS**            (A graph-theoretical index that reflects topological shape and size of amino acid side chains, *[15] C. Raychaudhury, A. Banerjee, P. Bag, S. Roy, J. Chem. Inf. Comput. Sci. 1999, 39, 248–254.*)
    - **MSW**            (Amino acid scale based on a pca of the molecular surface based WHIM descriptor (MS-WHIM), extended to natural amino acids, *[16] A. Zaliani, E. Gancia, J. Chem. Inf. Comput. Sci 1999, 39, 525–533.*)
    - **pepArc**         (modlabs pharmacophoric feature scale, dimensions are: hydrophobicity, polarity, positive charge, negative charge, proline.)
    - **pepcats**        (modlabs pharmacophoric feature based PEPCATS scale, *[17] C. P. Koch, A. M. Perna, M. Pillong, N. K. Todoroff, P. Wrede, G. Folkers, J. A. Hiss, G. Schneider, PLoS Comput. Biol. 2013, 9, e1003088.*)
    - **polarity**       (Amino acid polarity scale, *[18] J. M. Zimmerman, N. Eliezer, R. Simha, J. Theor. Biol. 1968, 21, 170–201.*)
    - **PPCALI**         (modlabs inhouse scale derived from a pca of 143 amino acid property scales, *[19] C. P. Koch, A. M. Perna, M. Pillong, N. K. Todoroff, P. Wrede, G. Folkers, J. A. Hiss, G. Schneider, PLoS Comput. Biol. 2013, 9, e1003088.*)
    - **refractivity**   (Relative amino acid refractivity values, *[20] T. L. McMeekin, M. Wilensky, M. L. Groves, Biochem. Biophys. Res. Commun. 1962, 7, 151–156.*)
    - **t_scale**        (A pca derived scale based on amino acid side chain properties calculated with 6 different probes of the GRID program, *[21] M. Cocchi, E. Johansson, Quant. Struct. Act. Relationships 1993, 12, 1–8.*)
    - **TM_tend**        (Amino acid transmembrane propensity scale, extracted from http://web.expasy.org/protscale, *[22] Zhao, G., London E. Protein Sci. 2006, 15, 1987-2001.*)
    - **z3**             (The original three dimensional Z-scale, *[23] S. Hellberg, M. Sjöström, B. Skagerberg, S. Wold, J. Med. Chem. 1987, 30, 1126–1135.*)
    - **z5**             (The extended five dimensional Z-scale, *[24] M. Sandberg, L. Eriksson, J. Jonsson, M. Sjöström, S. Wold, J. Med. Chem. 1998, 41, 2481–2491.*)

    Further, amino acid scale independent methods can be calculated with help of the :class:`GlobalDescriptor` class.

    """

    def __init__(self, seqs, scalename='Eisenberg'):
        """
        :param seqs: a .fasta file with sequences, a list of sequences or a single sequence as string to calculate the
            descriptor values for.
        :param scalename: {str} name of the amino acid scale (one of the given list above) used to calculate the
            descriptor values
        :return: initialized attributes :py:attr:`sequences`, :py:attr:`names` and dictionary :py:attr:`scale` with
            amino acid scale values of the scale name in :py:attr:`scalename`.
        :Example:

        >>> AMP = PeptideDescriptor('KLLKLLKKLLKLLK','pepcats')
        >>> AMP.sequences
        ['KLLKLLKKLLKLLK']
        >>> seqs = PeptideDescriptor('/Path/to/file.fasta', 'eisenberg')  # load sequences from .fasta file
        >>> seqs.sequences
        ['AFDGHLKI','KKLQRSDLLRTK','KKLASCNNIPPR'...]
        """
        super(PeptideDescriptor, self).__init__(seqs)
        self.scalename, self.scale = load_scale(scalename.lower())
        self.all_moms = list()  # for passing hydrophobic moments to calculate_profile
        self.all_globs = list()  # for passing global  to calculate_profile

    def load_scale(self, scalename):
        """Method to load amino acid values from a given scale

        :param scalename: {str} name of the amino acid scale to be loaded.
        :return: loaded amino acid scale values in a dictionary in the attribute :py:attr:`scale`.

        .. seealso:: :func:`modlamp.core.load_scale()`
        """
        self.scalename, self.scale = load_scale(scalename.lower())

    def calculate_autocorr(self, window, append=False):
        """Method for auto-correlating the amino acid values for a given descriptor scale

        :param window: {int} correlation window for descriptor calculation in a sliding window approach
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: calculated descriptor numpy.array in the attribute :py:attr:`descriptor`.
        :Example:

        >>> AMP = PeptideDescriptor('GLFDIVKKVVGALGSL','PPCALI')
        >>> AMP.calculate_autocorr(7)
        >>> AMP.descriptor
        array([[  1.28442339e+00,   1.29025116e+00,   1.03240901e+00, .... ]])
        >>> AMP.descriptor.shape
        (1, 133)

        .. versionchanged:: v.2.3.0
        """
        desc = Parallel(n_jobs=-1)(delayed(_one_autocorr)(seq, window, self.scale) for seq in self.sequences)

        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
        else:
            self.descriptor = np.array(desc)

    def calculate_crosscorr(self, window, append=False):
        """Method for cross-correlating the amino acid values for a given descriptor scale

        :param window: {int} correlation window for descriptor calculation in a sliding window approach
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: calculated descriptor numpy.array in the attribute :py:attr:`descriptor`.
        :Example:

        >>> AMP = PeptideDescriptor('GLFDIVKKVVGALGSL','pepcats')
        >>> AMP.calculate_crosscorr(7)
        >>> AMP.descriptor
        array([[ 0.6875    ,  0.46666667,  0.42857143,  0.61538462,  0.58333333, ... ]])
        >>> AMP.descriptor.shape
        (1, 147)
        """
        desc = Parallel(n_jobs=-1)(delayed(_one_crosscorr)(seq, window, self.scale) for seq in self.sequences)

        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
        else:
            self.descriptor = np.array(desc)

    def calculate_moment(self, window=1000, angle=100, modality='max', append=False):
        """Method for calculating the maximum or mean moment of the amino acid values for a given descriptor scale and
        window.

        :param window: {int} amino acid window in which to calculate the moment. If the sequence is shorter than the
            window, the length of the sequence is taken. So if the default window of 1000 is chosen, for all sequences
            shorter than 1000, the **global** hydrophobic moment will be calculated. Otherwise, the maximal
            hydrophiobic moment for the chosen window size found in the sequence will be returned.
        :param angle: {int} angle in which to calculate the moment. **100** for alpha helices, **180** for beta sheets.
        :param modality: {'all', 'max' or 'mean'} Calculate respectively maximum or mean hydrophobic moment. If all,
            moments for all windows are returned.
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: Calculated descriptor as a numpy.array in the attribute :py:attr:`descriptor` and all possible global
            values in :py:attr:`all_moms` (needed for the :py:func:`calculate_profile` method)
        :Example:

        >>> AMP = PeptideDescriptor('GLFDIVKKVVGALGSL', 'eisenberg')
        >>> AMP.calculate_moment()
        >>> AMP.descriptor
        array([[ 0.48790226]])
        """
        if self.scale['A'] == list:
            print('\n Descriptor moment calculation is only possible for one dimensional pydpi_py3.\n')
    
        else:
            desc = []
            for seq in self.sequences:
                wdw = min(window, len(seq))  # if sequence is shorter than window, take the whole sequence instead
                mtrx = []
                mwdw = []
                
                for aa in range(len(seq)):
                    mtrx.append(self.scale[str(seq[aa])])
                    
                for i in range(len(mtrx) - wdw + 1):
                    mwdw.append(sum(mtrx[i:i + wdw], []))
    
                mwdw = np.asarray(mwdw)
                rads = angle * (np.pi / 180) * np.asarray(range(wdw))  # calculate actual moment (radial)
                vcos = (mwdw * np.cos(rads)).sum(axis=1)
                vsin = (mwdw * np.sin(rads)).sum(axis=1)
                moms = np.sqrt(vsin ** 2 + vcos ** 2) / wdw
    
                if modality == 'max':  # take window with maximal value
                    moment = np.max(moms)
                elif modality == 'mean':  # take average value over all windows
                    moment = np.mean(moms)
                elif modality == 'all':
                    moment = moms
                else:
                    print('\nERROR!\nModality parameter is wrong, please choose between "all", "max" and "mean".\n')
                    return
                desc.append(moment)
                self.all_moms.append(moms)
                
            desc = np.asarray(desc).reshape(len(desc), 1)  # final descriptor array
            
            if append:
                self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            else:
                self.descriptor = np.array(desc)

    def calculate_global(self, window=1000, modality='max', append=False):
        """Method for calculating a global / window averaging descriptor value of a given AA scale

        :param window: {int} amino acid window in which to calculate the moment. If the sequence is shorter than the
            window, the length of the sequence is taken.
        :param modality: {'max' or 'mean'} Calculate respectively maximum or mean hydrophobic moment.
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: Calculated descriptor as a numpy.array in the attribute :py:attr:`descriptor` and all possible global
            values in :py:attr:`all_globs` (needed for the :py:func:`calculate_profile` method)
        :Example:

        >>> AMP = PeptideDescriptor('GLFDIVKKVVGALGSL','eisenberg')
        >>> AMP.calculate_global(window=1000, modality='max')
        >>> AMP.descriptor
        array([[ 0.44875]])
        """
        desc = list()
        for n, seq in enumerate(self.sequences):
            wdw = min(window, len(seq))  # if sequence is shorter than window, take the whole sequence instead
            mtrx = []
            mwdw = []
            
            for l in range(len(seq)):  # translate AA sequence into values
                mtrx.append(self.scale[str(seq[l])])

            for i in range(len(mtrx) - wdw + 1):
                mwdw.append(sum(mtrx[i:i + wdw], []))  # list of all the values for the different windows
                
            mwdw = np.asarray(mwdw)
            glob = np.sum(mwdw, axis=1) / float(wdw)
            outglob = float()
            
            if modality in ['max', 'mean']:
                if modality == 'max':
                    outglob = np.max(glob)  # returned moment will be the maximum of all windows
                elif modality == 'mean':
                    outglob = np.mean(glob)  # returned moment will be the mean of all windows
                else:
                    print('Modality parameter is wrong, please choose between "max" and "mean"\n.')
                    return
            desc.append(outglob)
            self.all_globs.append(glob)
        
        desc = np.asarray(desc).reshape(len(desc), 1)
        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
        else:
            self.descriptor = np.array(desc)

    def calculate_profile(self, prof_type='uH', window=7, append=False):
        """Method for calculating hydrophobicity or hydrophobic moment profiles for given sequences and fitting for
        slope and intercept. The hydrophobicity scale used is "eisenberg"

        :param prof_type: prof_type of profile, available: 'H' for hydrophobicity or 'uH' for hydrophobic moment
        :param window: {int} size of sliding window used (odd-numbered).
        :param append: {boolean} whether the produced descriptor values should be appended to the existing ones in the
            attribute :py:attr:`descriptor`.
        :return: Fitted slope and intercept of calculated profile for every given sequence in the attribute
            :py:attr:`descriptor`.
        :Example:

        >>> AMP = PeptideDescriptor('KLLKLLKKVVGALG','kytedoolittle')
        >>> AMP.calculate_profile(prof_type='H')
        >>> AMP.descriptor
        array([[ 0.03731293,  0.19246599]])
        """
        if prof_type == 'uH':
            self.calculate_moment(window=window)
            y_vals = self.all_moms
        elif prof_type == 'H':
            self.calculate_global(window=window)
            y_vals = self.all_globs
        else:
            print('prof_type parameter is unknown, choose "uH" for hydrophobic moment or "H" for hydrophobicity\n.')
            sys.exit()

        desc = list()
        for n, seq in enumerate(self.sequences):
            x_vals = range(len(seq))[int((window - 1) / 2):-int((window - 1) / 2)]
            if len(seq) <= window:
                slope, intercept, r_value, p_value, std_err = [0, 0, 0, 0, 0]
            else:
                slope, intercept, r_value, p_value, std_err = stats.linregress(x_vals, y_vals[n])
            desc.append([slope, intercept])

        if append:
            self.descriptor = np.hstack((self.descriptor, np.array(desc)))
        else:
            self.descriptor = np.array(desc)

    def calculate_arc(self, modality="max", append=False):
        """ Method for calculating property arcs as seen in the helical wheel plot. Use for binary amino acid scales only.
        
        :param modality: modality of the arc to calculate, to choose between "max" and "mean".
        :param append: if true, append to current descriptor stored in the descriptor attribute.
        :return: calculated descriptor as numpy.array in the descriptor attribute.

        :Example:

        >>> arc = PeptideDescriptor("KLLKLLKKLLKLLK", scalename="peparc")
        >>> arc.calculate_arc(modality="max", append=False)
        >>> arc.descriptor
        array([[200, 160, 160,   0,   0]])
        """
        desc = Parallel(n_jobs=-1)(delayed(_one_arc)(seq, modality, self.scale) for seq in self.sequences)

        # Converts each of the amino acids to descriptor vector
        for seq in self.sequences:

            # desc_mat = []
            # for aa in seq:
            #     desc_mat.append(self.scale[aa])
            # desc_mat = np.asarray(desc_mat)
			#
            # # Check descriptor dimension
            # desc_dim = desc_mat.shape[1]
			#
            # # list to store descriptor values for all windows
            # allwindows_arc = []
			#
            # if len(seq) > 18:
            #     window = 18
            #     # calculates number of windows in sequence
            #     num_windows = len(seq) - window
            # else:
            #     window = len(seq)
            #     num_windows = 1
			#
            # # loop through all windows
            # for j in range(num_windows):
            #     # slices descriptor matrix into current window
            #     window_mat = desc_mat[j:j + window, :]
			#
            #     # defines order of amino acids in helical projection
            #     order = [0, 11, 4, 15, 8, 1, 12, 5, 16, 9, 2, 13, 6, 17, 10, 3, 14, 7]
			#
            #     # orders window descriptor matrix into helical projection order
            #     ordered = []
            #     for pos in order:
            #         try:
            #             ordered.append(window_mat[pos, :])
            #         except:
            #             # for sequences of len < 18 adding dummy vector with 2s, length of descriptor dimensions
            #             ordered.append([2] * desc_dim)
            #     ordered = np.asarray(ordered)
			#
            #     window_arc = []
			#
            #     # loop through pharmacophoric features
            #     for m in range(desc_dim):
            #         all_arcs = []  # stores all arcs that can be found of a pharmacophoric feature
            #         arc = 0
			#
            #         for n in range(18):  # for all positions in helix, regardless of sequence length
            #             if ordered[n, m] == 0:  # if position does not contain pharmacophoric feature
            #                 all_arcs.append(arc)  # append previous arc to all arcs list
            #                 arc = 0  # arc is initialized
            #             elif ordered[n, m] == 1:  # if position contains pharmacophoric feature(PF), elongate arc by 20°
            #                 arc += 20
            #             elif ordered[n, m] == 2:  # if position doesn't contain amino acid:
            #                 if ordered[n - 1, m] == 1:  # if previous position contained PF add 10°
            #                     arc += 10
            #                 elif ordered[n - 1, m] == 0:  # if previous position didn't contain PF don't add anything
            #                     arc += 0
            #                 elif ordered[
            #                             n - 2, m] == 1:  # if previous position is empty then check second previous for PF
            #                     arc += 10
            #                 if n == 17:  # if we are at the last position check for position n=0 instead of next position.
            #                     if ordered[0, m] == 1:  # if it contains PF add 10° extra
            #                         arc += 10
            #                 else:  # if next position contains PF add 10° extra
            #                     if ordered[n + 1, m] == 1:
            #                         arc += 10
            #                     elif ordered[n + 1, m] == 0:
            #                         arc += 0
            #                     else:  # if next position is empty check for 2nd next position
            #                         if n == 16:
            #                             if ordered[0, m] == 1:
            #                                 arc += 10
            #                         else:
            #                             if ordered[n + 2, m] == 1:
            #                                 arc += 10
			#
            #         all_arcs.append(arc)
            #         if not arc == 360:
            #             arc0 = all_arcs.pop() + all_arcs[0]  # join first and last arc together
            #             all_arcs = [arc0] + all_arcs[1:]
			#
            #         window_arc.append(np.max(all_arcs))  # append to window arcs the maximum arc of this PF
            #     allwindows_arc.append(window_arc)  # append all PF arcs of this window
			#
            # allwindows_arc = np.asarray(allwindows_arc)
			#
            # if modality == 'max':
            #     final_arc = np.max(allwindows_arc, axis=0)  # calculate maximum / mean arc along all windows
            # elif modality == 'mean':
            #     final_arc = np.mean(allwindows_arc, axis=0)
            # else:
            #     print('modality is unknown, please choose between "max" and "mean"\n.')
            #     sys.exit()

            if append:
                self.descriptor = np.hstack((self.descriptor, np.array(desc)))
            else:
                self.descriptor = np.array(desc)











import pandas as pd
from typing import List
from .descriptors import DNADescriptor

def _calculate_descriptors(data: pd.DataFrame, descriptor_list: List) -> pd.DataFrame:
    """
    From a dataset of sequences and labels, this function calculates the descriptors and returns a dataframe with them.
    The user can also specify which descriptors to calculate.
    """
    list_feature = []
    count = 0
    for seq in data['sequence']:
        res = {'sequence': seq}
        dna = DNADescriptor(seq)
        features = dna.get_descriptors(descriptor_list)
        res.update(features)
        list_feature.append(res)

        # print progress every 100 sequences
        if count % 100 == 0:
            print(count, '/', len(data))

        count += 1
    print("Done!")
    df = pd.DataFrame(list_feature)
    return df


def _process_lists(fps_x, field):
    """
    A helper function to normalize lists.
    """
    l = fps_x[field].to_list()
    new_df = pd.DataFrame(l)
    new_df.columns = [str(field) + "_" + str(i) for i in new_df.columns]
    fps_x.drop(field, axis=1, inplace=True)
    return new_df


def _process_lists_of_lists(fps_x, field):
    """
    A helper function to normalize lists of lists.
    """
    l = fps_x[field].to_list()
    new_df = pd.DataFrame(l)
    new_df.columns = [str(field) + "_" + str(i) for i in new_df.columns]
    empty_val = {} if field == "enhanced_nucleic_acid_composition" else []
    small_processed = []
    for f in new_df.columns:
        col = [empty_val if i is None else i for i in new_df[f].to_list()]
        sub = pd.DataFrame(col)
        sub.columns = [str(f) + "_" + str(i) for i in sub.columns]
        small_processed.append(sub)
    fps_x.drop(field, axis=1, inplace=True)
    return small_processed



def normalization(fps_x, descriptor_list):
    """
    Because the model cannot process data in dictionaries and lists, the descriptors that produce these forms must still be normalized.

    To normalize the data, dicts and lists need to "explode" into more columns. 

    E.g. dicts:

    | descriptor_hello |
    | ---------------- |
    | {'a': 1, 'b': 2} |

    will be transformed into:

    | descriptor_hello_a | descriptor_hello_b |
    | ------------------ | ------------------ |
    | 1                  | 2                  |

    E.g. lists:

    | descriptor_hello |
    | ---------------- |
    | [1, 2, 3]        |

    will be transformed into:

    | descriptor_hello_0 | descriptor_hello_1 | descriptor_hello_2 |
    | ------------------ | ------------------ | ------------------ |
    | 1                  | 2                  | 3                  |
    """
    lists = ["nucleic_acid_composition", "dinucleotide_composition", "trinucleotide_composition",
             "k_spaced_nucleic_acid_pairs", "kmer", "PseDNC", "PseKNC", "DAC", "DCC", "DACC", "TAC", "TCC", "TACC"]
    lists_of_lists = [
        "accumulated_nucleotide_frequency"
    ]
    
    # update to be normalized lists with only columns the user wants
    if(descriptor_list != []):
        lists = [l for l in lists if l in descriptor_list]
        lists_of_lists = [l for l in lists_of_lists if l in descriptor_list]

    small_processed = []
    for i in lists:
        new_df = _process_lists(fps_x, i)
        small_processed.append(new_df)

    for i in lists_of_lists:
        smaller_processed = _process_lists_of_lists(fps_x, i)
        small_processed += smaller_processed

    new_fps_x = pd.concat([fps_x, *small_processed], axis=1)
    return new_fps_x


def calculate_and_normalize(data: pd.DataFrame, descriptor_list: list = []) -> pd.DataFrame:
    """
    This function calculates the descriptors and normalizes the data all at once from a dataframe of sequences and labels. The user can also specify which descriptors to calculate.
    """
    features = _calculate_descriptors(data, descriptor_list)
    if 'label' in data:
        fps_y = data['label']
    else:
        fps_y = None
    fps_x = features.loc[:, features.columns != 'label']
    fps_x = fps_x.loc[:, fps_x.columns != 'sequence']
    fps_x = normalization(fps_x, descriptor_list)
    return fps_x, fps_y

if __name__ == "__main__":
    from sequence import ReadDNA
    reader = ReadDNA()
    filename = 'datasets/primer/dataset.csv'
    data = reader.read_csv(filename=filename, with_labels=True)
    fps_x, fps_y = calculate_and_normalize(data)
    print(fps_x)


import os
import pickle
import random
import sys
import math
import numpy as np
from itertools import product
import json

ALPHABET = 'ACGT'
ALPHABET_CUT = 'ACGTN'
pairs = {
    'A': 'T',
    'T': 'A',
    'G': 'C',
    'C': 'G'
}

# -----------------------------------------------------------------------------
def seed_everything(seed=42):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ["CUBLAS_WORKSPACE_CONFIG"]=":4096:2"
    np.random.seed(seed)
    # torch.manual_seed(seed)
    # torch.cuda.manual_seed(seed)
    # torch.cuda.manual_seed_all(seed)
    # torch.backends.cudnn.deterministic = True
    # torch.backends.cudnn.benchmark = False
    # torch.backends.cudnn.enabled = False

def checker(sequence):
    """
    Checks if the input sequence is a valid DNA sequence.
    """
    return all(i in ALPHABET for i in sequence)

def checker_cut(sequence):
    """
    Checks if the input sequence is a valid DNA sequence. Includes the 'N' character as valid because it is used to fill the sequence to the right length.
    """
    return all(i in ALPHABET_CUT for i in sequence)

def normal_round(n):
    """
    Equivalent to python's round but its rounds up if it ends up with 0.5.
    """
    if n - math.floor(n) < 0.5:
        return math.floor(n)
    return math.ceil(n)


def normalize_dict(dic):
    """Normalize the value of a dictionary."""
    N = sum(dic.values())
    for key in dic:
        dic[key] = round(dic[key] / N, 3)
    return dic


def make_kmer_list(k):
    try:
        return ["".join(e) for e in product(ALPHABET, repeat=k)]
    except TypeError:
        print("TypeError: k must be an inter and larger than 0, alphabet must be a string.")
        raise TypeError
    except ValueError:
        print("TypeError: k must be an inter and larger than 0")
        raise ValueError


def make_kmer_dict(k):
    try:
        return {''.join(i): 0 for i in product(ALPHABET, repeat=k)}
    except TypeError:
        print("TypeError: k must be an inter and larger than 0, alphabet must be a string.")
        raise TypeError
    except ValueError:
        print("TypeError: k must be an inter and larger than 0")
        raise ValueError

def calculate_kmer_onehot(k):
    nucleotides = [''.join(i) for i in product(ALPHABET_CUT, repeat=k)]
    encoded = []
    for i in range(5 ** k):
        encoded.append(np.zeros(5 ** k).tolist())
        encoded[i][i] = 1.0

    return {nucleotides[i]: encoded[i] for i in range(len(nucleotides))}

def calculate_kmer_list(sequence, k):
    l = []
    for i in range(len(sequence) - k + 1):
        l.append(sequence[i:i+k])
    return l

def read_config(device, filename='config.json'):
    """
    Reads the configuration file and validates the values. Returns the configuration.
    """
    with open(filename) as f:
        config = json.load(f)

    # ------------------------------------ check if data_dir exists ------------------------------------
    current_path = os.getcwd()
    current_path = current_path.replace("/notebooks", "") # when running from notebook

    config['combination']['data_dir'] = current_path + '/datasets/' + config['combination']['data_dir']
    if not os.path.exists(config['combination']['data_dir']):
        raise ValueError("Data directory does not exist:", config['combination']['data_dir'])

        # --------------------------- check if model and mode combination is valid -------------------------
    model_label = config['combination']['model_label']
    mode = config['combination']['mode']
    if(model_label in combinations):
        if(mode not in combinations[model_label]):
            raise ValueError(model_label, 'does not support', mode, ', please choose one of', combinations[model_label])
    else:
        raise ValueError('Model label:', model_label, 'not implemented in', combinations.keys())

    # --------------------------- check if it's binary classification ----------------------------------
    loss = config['fixed_vals']['loss_function']
    output_size = config['fixed_vals']['output_size']
    if(loss != "cross_entropy" or output_size != 2):
        raise ValueError(
            'Model is not binary classification, please set loss_function to cross_entropy and output_size to 2')

    # --------------------------- create the cross entropy pytorch object ------------------------------
    class_weights = torch.tensor(config['combination']['class_weights']).to(device)
    config['fixed_vals']['loss_function'] = nn.CrossEntropyLoss(weight=class_weights)

    # --------------------------- create ray tune objects ----------------------------------------------
    config['hyperparameter_search_space']["hidden_size"] = tune.choice(config['hyperparameter_search_space']['hidden_size'])
    config['hyperparameter_search_space']["lr"] = tune.choice(config['hyperparameter_search_space']['lr'])
    config['hyperparameter_search_space']["batch_size"] = tune.choice(config['hyperparameter_search_space']['batch_size'])
    config['hyperparameter_search_space']["dropout"] = tune.choice(config['hyperparameter_search_space']['dropout'])

    if config['combination']['model_label'] not in ['mlp', 'cnn']:
        config['hyperparameter_search_space']["num_layers"] = tune.choice(config['hyperparameter_search_space']['num_layers'])

    return config

# ----------------------------------------------------------------------------------------------------
# ---------------------------------- Auxiliary function for models -----------------------------------
# ----------------------------------------------------------------------------------------------------

def calc_maxpool_output(hidden_size, sequence_length):
    conv1_padding = 0
    conv1_dilation = 1
    conv1_kernel_size = 12
    conv1_stride = 1

    l_out = ((sequence_length + 2*conv1_padding - conv1_dilation*(conv1_kernel_size-1) - 1)/conv1_stride + 1)
    maxpool_padding = 0
    maxpool_dilation = 1
    maxpool_stride = 5
    maxpool_kernel_size = 12
    max_pool_output = int((l_out+2*maxpool_padding-maxpool_dilation*(maxpool_kernel_size-1)-1)/maxpool_stride+1)

    max_pool_output *= hidden_size

    return max_pool_output


# ----------------------------------------------------------------------------------------------------
# ------------------- The following functions were retrieved from repDNA package ---------------------
# ----------------------------------------------------------------------------------------------------

def ready_acc(k, phyche_index=None, all_property=False, extra_phyche_index=None):
    """Public function for get sequence_list and phyche_value.
    """
    if phyche_index is None:
        phyche_index = []
    if extra_phyche_index is None:
        extra_phyche_index = {}
    phyche_value = generate_phyche_value(k, phyche_index, all_property, extra_phyche_index)

    return phyche_value


def generate_phyche_value(k, phyche_index=None, all_property=False, extra_phyche_index=None):
    """Combine the user selected phyche_list, is_all_property and extra_phyche_index to a new standard phyche_value."""
    if phyche_index is None:
        phyche_index = []
    if extra_phyche_index is None:
        extra_phyche_index = {}

    diphyche_list = ['Base stacking', 'Protein induced deformability', 'B-DNA twist', 'Dinucleotide GC Content',
                     'A-philicity', 'Propeller twist', 'Duplex stability:(freeenergy)',
                     'Duplex tability(disruptenergy)', 'DNA denaturation', 'Bending stiffness', 'Protein DNA twist',
                     'Stabilising energy of Z-DNA', 'Aida_BA_transition', 'Breslauer_dG', 'Breslauer_dH',
                     'Breslauer_dS', 'Electron_interaction', 'Hartman_trans_free_energy', 'Helix-Coil_transition',
                     'Ivanov_BA_transition', 'Lisser_BZ_transition', 'Polar_interaction', 'SantaLucia_dG',
                     'SantaLucia_dH', 'SantaLucia_dS', 'Sarai_flexibility', 'Stability', 'Stacking_energy',
                     'Sugimoto_dG', 'Sugimoto_dH', 'Sugimoto_dS', 'Watson-Crick_interaction', 'Twist', 'Tilt',
                     'Roll', 'Shift', 'Slide', 'Rise']
    triphyche_list = ['Dnase I', 'Bendability (DNAse)', 'Bendability (consensus)', 'Trinucleotide GC Content',
                      'Nucleosome positioning', 'Consensus_roll', 'Consensus-Rigid', 'Dnase I-Rigid', 'MW-Daltons',
                      'MW-kg', 'Nucleosome', 'Nucleosome-Rigid']

    # Set and check physicochemical properties.
    if 2 == k:
        if all_property is True:
            phyche_index = diphyche_list
        else:
            for e in phyche_index:
                if e not in diphyche_list:
                    raise ValueError(" ".join(["Sorry, the physicochemical properties", e, "is not exit."]))
    elif 3 == k:
        if all_property is True:
            phyche_index = triphyche_list
        else:
            for e in phyche_index:
                if e not in triphyche_list:
                    raise ValueError(" ".join(["Sorry, the physicochemical properties", e, "is not exit."]))

    return extend_phyche_index(get_phyche_index(k, phyche_index), extra_phyche_index)


def get_phyche_index(k, phyche_list):
    """get phyche_value according phyche_list."""
    phyche_value = {}
    if 0 == len(phyche_list):
        for nucleotide in make_kmer_list(k):
            phyche_value[nucleotide] = []
        return phyche_value

    nucleotide_phyche_value = get_phyche_factor_dic(k)
    for nucleotide in make_kmer_list(k):
        if nucleotide not in phyche_value:
            phyche_value[nucleotide] = []
        for e in nucleotide_phyche_value[nucleotide]:
            if e[0] in phyche_list:
                phyche_value[nucleotide].append(e[1])

    return phyche_value


def extend_phyche_index(original_index, extend_index):
    """Extend {phyche:[value, ... ]}"""
    if extend_index is None or len(extend_index) == 0:
        return original_index
    for key in list(original_index.keys()):
        original_index[key].extend(extend_index[key])
    return original_index


def get_phyche_factor_dic(k):
    """Get all {nucleotide: [(phyche, value), ...]} dict."""
    full_path = os.path.realpath(__file__)
    if 2 == k:
        file_path = "%s/data/mmc3.data" % os.path.dirname(full_path)
    elif 3 == k:
        file_path = "%s/data/mmc4.data" % os.path.dirname(full_path)
    else:
        sys.stderr.write("The k can just be 2 or 3.")
        sys.exit(0)

    try:
        with open(file_path, 'rb') as f:
            phyche_factor_dic = pickle.load(f)
    except:
        with open(file_path, 'r') as f:
            phyche_factor_dic = pickle.load(f)

    return phyche_factor_dic


def make_ac_vector(sequence_list, lag, phyche_value, k):
    phyche_values = list(phyche_value.values())
    len_phyche_value = len(phyche_values[0])

    vec_ac = []
    for sequence in sequence_list:
        len_seq = len(sequence)
        each_vec = []

        for temp_lag in range(1, lag + 1):
            for j in range(len_phyche_value):

                # Calculate average phyche_value for a nucleotide.
                ave_phyche_value = 0.0
                for i in range(len_seq - temp_lag - k + 1):
                    nucleotide = sequence[i: i + k]
                    ave_phyche_value += float(phyche_value[nucleotide][j])
                ave_phyche_value /= len_seq

                # Calculate the vector.
                temp_sum = 0.0
                for i in range(len_seq - temp_lag - k + 1):
                    nucleotide1 = sequence[i: i + k]
                    nucleotide2 = sequence[i + temp_lag: i + temp_lag + k]
                    temp_sum += (float(phyche_value[nucleotide1][j]) - ave_phyche_value) * (
                        float(phyche_value[nucleotide2][j]))

                try:
                    val = round(temp_sum / (len_seq - temp_lag - k + 1), 3)
                except ZeroDivisionError:
                    val = 0.0
                each_vec.append(val)
        vec_ac.append(each_vec)

    return vec_ac


def make_cc_vector(sequence_list, lag, phyche_value, k):
    phyche_values = list(phyche_value.values())
    len_phyche_value = len(phyche_values[0])

    vec_cc = []
    for sequence in sequence_list:
        len_seq = len(sequence)
        each_vec = []

        for temp_lag in range(1, lag + 1):
            for i1 in range(len_phyche_value):
                for i2 in range(len_phyche_value):
                    if i1 != i2:
                        # Calculate average phyche_value for a nucleotide.
                        ave_phyche_value1 = 0.0
                        ave_phyche_value2 = 0.0
                        for j in range(len_seq - temp_lag - k + 1):
                            nucleotide = sequence[j: j + k]
                            ave_phyche_value1 += float(phyche_value[nucleotide][i1])
                            ave_phyche_value2 += float(phyche_value[nucleotide][i2])
                        ave_phyche_value1 /= len_seq
                        ave_phyche_value2 /= len_seq

                        # Calculate the vector.
                        temp_sum = 0.0
                        for j in range(len_seq - temp_lag - k + 1):
                            nucleotide1 = sequence[j: j + k]
                            nucleotide2 = sequence[j + temp_lag: j + temp_lag + k]
                            temp_sum += (float(phyche_value[nucleotide1][i1]) - ave_phyche_value1) * \
                                        (float(phyche_value[nucleotide2][i2]) - ave_phyche_value2)
                        try:
                            val = round(temp_sum / (len_seq - temp_lag - k + 1), 3)
                        except ZeroDivisionError:
                            val = 0.0
                        each_vec.append(val)

        vec_cc.append(each_vec)

    return vec_cc
